"""Tests for dataset utils."""

from typing import Iterable, Iterator

from ..schema import PathTuple
from ..utils import chunks
from .dataset_utils import count_leafs, sparse_to_dense_compute, wrap_in_dicts


def test_count_nested() -> None:
  a = [[1, 2], [[3]], [4, 5, 6]]
  assert 6 == count_leafs(a)


def test_wrap_in_dicts_with_spec_of_one_repeated() -> None:
  a = [[1, 2], [3], [4, 5, 5]]
  spec: list[PathTuple] = [('a', 'b', 'c'), ('d',)]  # Corresponds to a.b.c.*.d.
  result = wrap_in_dicts(a, spec)
  assert list(result) == [
    {'a': {'b': {'c': [{'d': 1}, {'d': 2}]}}},
    {'a': {'b': {'c': [{'d': 3}]}}},
    {'a': {'b': {'c': [{'d': 4}, {'d': 5}, {'d': 5}]}}},
  ]


def test_wrap_in_dicts_with_spec_of_double_repeated() -> None:
  a = [[[1, 2], [3, 4, 5]], [[6]], [[7], [8], [9, 10]]]
  spec: list[PathTuple] = [('a', 'b'), tuple(), ('c',)]  # Corresponds to a.b.*.*.c.
  result = wrap_in_dicts(a, spec)
  assert list(result) == [
    {'a': {'b': [[{'c': 1}, {'c': 2}], [{'c': 3}, {'c': 4}, {'c': 5}]]}},
    {'a': {'b': [[{'c': 6}]]}},
    {'a': {'b': [[{'c': 7}], [{'c': 8}], [{'c': 9}, {'c': 10}]]}},
  ]


def test_sparse_to_dense_compute() -> None:
  sparse_input = iter([None, 1, 7, None, None, 3, None, 5, None, None])

  def func(xs: Iterable[int]) -> Iterator[int]:
    for x in xs:
      yield x + 1

  out = sparse_to_dense_compute(sparse_input, func)
  assert list(out) == [None, 2, 8, None, None, 4, None, 6, None, None]


def test_sparse_to_dense_compute_batching() -> None:
  sparse_input = iter([None, 1, 7, None, None, 3, None, 5, None, None])

  def func(xs: Iterable[int]) -> Iterator[int]:
    for batch in chunks(xs, 2):
      yield batch[0] + 1
      if len(batch) > 1:
        yield batch[1] + 1

  out = sparse_to_dense_compute(sparse_input, func)
  assert list(out) == [None, 2, 8, None, None, 4, None, 6, None, None]


def test_fully_dense() -> None:
  sparse_input = iter([1, 7, 3, 5])

  def func(xs: Iterable[int]) -> Iterator[int]:
    for x in xs:
      yield x + 1

  out = sparse_to_dense_compute(sparse_input, func)
  assert list(out) == [2, 8, 4, 6]


def test_sparse_to_dense_compute_fully_sparse() -> None:
  sparse_input = iter([None, None, None])

  def func(xs: Iterable[int]) -> Iterator[int]:
    for x in xs:
      yield x + 1

  out = sparse_to_dense_compute(sparse_input, func)
  assert list(out) == [None, None, None]


def test_sparse_to_dense_compute_empty() -> None:
  sparse_input: Iterator[int] = iter([])

  def func(xs: Iterable[int]) -> Iterator[int]:
    for x in xs:
      yield x + 1

  out = sparse_to_dense_compute(sparse_input, func)
  assert list(out) == []


def test_sparse_to_dense_compute_iterable_returned() -> None:
  sparse_input: Iterator[int] = iter([1, 2, 3])

  # Users will sometimes write a Signal/Map function that's supposed to return an iterator
  # but accidentally return an iterable. This test ensures that we are robust to this mistake.
  def func(xs: Iterable[int]) -> Iterator[int]:
    return [x + 1 for x in xs]  # type: ignore

  out = sparse_to_dense_compute(sparse_input, func)
  assert list(out) == [2, 3, 4]



"""Test the PII signal."""


from ..schema import Item
from ..splitters.text_splitter_test_utils import text_to_expected_spans
from .pii import PII_CATEGORIES, SECRETS_KEY, PIISignal


def make_pii_dict(entries: dict[str, list[Item]]) -> dict[str, list[Item]]:
  """Make a PII dictionary with the given kwargs."""
  return {**{cat: [] for cat in list(PII_CATEGORIES.values()) + [SECRETS_KEY]}, **entries}


def test_pii_fields() -> None:
  signal = PIISignal()
  fields = signal.fields().fields
  assert fields is not None
  assert fields.keys() == set(list(PII_CATEGORIES.values()) + [SECRETS_KEY])


def test_pii_compute() -> None:
  signal = PIISignal()

  text = 'This is an email nik@test.com. pii@gmail.com are where emails are read.'
  emails = list(signal.compute([text]))

  expected_spans = text_to_expected_spans(text, ['nik@test.com', 'pii@gmail.com'])

  assert emails == [make_pii_dict({'email_address': expected_spans})]


def test_pii_case_insensitive() -> None:
  signal = PIISignal()

  text = 'These are some emails: NIK@Test.com. pII@gmAIL.COM are where emails are read.'
  emails = list(signal.compute([text]))

  expected_spans = text_to_expected_spans(text, ['NIK@Test.com', 'pII@gmAIL.COM'])

  assert emails == [make_pii_dict({'email_address': expected_spans})]


def test_ip_addresses() -> None:
  signal = PIISignal()

  text = 'These are some ip addresses: 192.158.1.38 and 2001:db8:3333:4444:5555:6666:7777:8888'
  pii = list(signal.compute([text]))
  expected_spans = text_to_expected_spans(
    text, ['192.158.1.38', '2001:db8:3333:4444:5555:6666:7777:8888']
  )
  assert pii == [make_pii_dict({'ip_address': expected_spans})]


def test_secrets() -> None:
  signal = PIISignal()

  text = 'These are some secrets: AKIATESTTESTTESTTEST'
  pii = list(signal.compute([text]))
  expected_spans = text_to_expected_spans(text, ['AKIATESTTESTTESTTEST'])
  assert pii == [make_pii_dict({SECRETS_KEY: expected_spans})]


def test_no_unicode_error() -> None:
  signal = PIISignal()

  text = 'Olof Heden and Denis S. Krotov \u2217'
  pii = list(signal.compute([text]))
  assert pii == [make_pii_dict({})]



"""Tests for the openchat format."""


from ..data.dataset_test_utils import TestDataMaker
from .openchat import OpenChat


def test_infer_open_chat(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [
      {
        'items': [
          {'role': 'user', 'content': 'Hello', 'weight': 0.0},
          {'role': 'assistant', 'content': 'Hi', 'weight': 1.0},
          {'role': 'user', 'content': 'How are you today?', 'weight': 0.0},
          {'role': 'assistant', 'content': "I'm fine.", 'weight': 1.0},
        ],
        'system': '',
      },
      {
        'items': [
          {'role': 'user', 'content': 'Who are you?', 'weight': 0.0},
          {'role': 'assistant', 'content': "I'm OpenChat.", 'weight': 1.0},
        ],
        'system': 'You are a helpful assistant named OpenChat.',
      },
    ]
  )

  assert dataset.manifest().dataset_format == OpenChat()



"""Manages mapping the dataset name to the database instance."""
import os
import pathlib
import threading
from typing import Optional, Type, Union

from pydantic import BaseModel

from .config import get_dataset_config
from .data.dataset import Dataset
from .env import get_project_dir
from .project import read_project_config
from .utils import get_datasets_dir

_DEFAULT_DATASET_CLS: Type[Dataset]

_CACHED_DATASETS: dict[str, Dataset] = {}

_db_lock = threading.Lock()


def _dataset_cache_key(
  namespace: str, dataset_name: str, project_dir: Union[str, pathlib.Path]
) -> str:
  """Get the cache key for a dataset."""
  return f'{os.path.abspath(project_dir)}/{namespace}/{dataset_name}'


def get_dataset(
  namespace: str, dataset_name: str, project_dir: Optional[Union[str, pathlib.Path]] = None
) -> Dataset:
  """Get the dataset instance."""
  if not _DEFAULT_DATASET_CLS:
    raise ValueError('Default dataset class not set.')

  project_dir = project_dir or get_project_dir()

  cache_key = _dataset_cache_key(namespace, dataset_name, project_dir)
  # https://docs.pytest.org/en/latest/example/simple.html#pytest-current-test-environment-variable
  inside_test = 'PYTEST_CURRENT_TEST' in os.environ
  with _db_lock:
    if cache_key not in _CACHED_DATASETS or inside_test:
      _CACHED_DATASETS[cache_key] = _DEFAULT_DATASET_CLS(
        namespace=namespace, dataset_name=dataset_name, project_dir=project_dir
      )
    return _CACHED_DATASETS[cache_key]


def remove_dataset_from_cache(
  namespace: str, dataset_name: str, project_dir: Optional[Union[str, pathlib.Path]] = None
) -> None:
  """Remove the dataset from the db manager cache."""
  project_dir = project_dir or get_project_dir()

  cache_key = _dataset_cache_key(namespace, dataset_name, project_dir)
  with _db_lock:
    if cache_key in _CACHED_DATASETS:
      del _CACHED_DATASETS[cache_key]


class DatasetInfo(BaseModel):
  """Information about a dataset."""

  namespace: str
  dataset_name: str
  description: Optional[str] = None
  tags: list[str] = []


def list_datasets(project_dir: Optional[Union[str, pathlib.Path]] = None) -> list[DatasetInfo]:
  """List the datasets in a project directory."""
  project_dir = project_dir or get_project_dir()

  datasets_path = get_datasets_dir(project_dir)

  # Skip if 'datasets' doesn't exist.
  if not os.path.isdir(datasets_path):
    return []

  project_config = read_project_config(project_dir)

  dataset_infos: list[DatasetInfo] = []
  for namespace in os.listdir(datasets_path):
    dataset_dir = os.path.join(datasets_path, namespace)
    # Skip if namespace is not a directory.
    if not os.path.isdir(dataset_dir):
      continue
    if namespace.startswith('.'):
      continue

    for dataset_name in os.listdir(dataset_dir):
      # Skip if dataset_name is not a directory.
      dataset_path = os.path.join(dataset_dir, dataset_name)
      if not os.path.isdir(dataset_path):
        continue
      if dataset_name.startswith('.'):
        continue

      dataset_config = get_dataset_config(project_config, namespace, dataset_name)
      tags: list[str] = []
      if dataset_config and dataset_config.settings and dataset_config.settings.tags:
        tags = dataset_config.settings.tags

      dataset_infos.append(DatasetInfo(namespace=namespace, dataset_name=dataset_name, tags=tags))

  return dataset_infos


# TODO(nsthorat): Make this a registry once we have multiple dataset implementations. This breaks a
# circular dependency.
def set_default_dataset_cls(dataset_cls: Type[Dataset]) -> None:
  """Set the default dataset class."""
  global _DEFAULT_DATASET_CLS
  _DEFAULT_DATASET_CLS = dataset_cls



"""PaLM embeddings."""
from typing import ClassVar, Optional

import numpy as np
from tenacity import retry, stop_after_attempt, wait_fixed, wait_random_exponential
from typing_extensions import override

from ..env import env
from ..schema import Item
from ..signal import TextEmbeddingSignal
from ..splitters.spacy_splitter import clustering_spacy_chunker
from ..tasks import TaskExecutionType
from .embedding import chunked_compute_embedding

PALM_BATCH_SIZE = 1  # PaLM API only supports batch size 1.
API_NUM_PARALLEL_REQUESTS = 256  # Because batch size is 1, we can send many requests in parallel.
GCP_NUM_PARALLEL_REQUESTS = 3  # We aim for 600 requests per minute
API_EMBEDDING_MODEL = 'models/embedding-gecko-001'
GCP_EMBEDDING_MODEL = 'textembedding-gecko@001'


class PaLM(TextEmbeddingSignal):
  """Computes embeddings using PaLM's embedding API.

  <br>**Important**: This will send data to an external server!

  <br>To use this signal, you must get a PaLM API key from
  [makersuite.google.com](https://makersuite.google.com/app/apikey) and add it to your .env.local.
  """

  name: ClassVar[str] = 'palm'
  display_name: ClassVar[str] = 'PaLM Embeddings'
  local_batch_size: ClassVar[int] = PALM_BATCH_SIZE
  local_parallelism: ClassVar[int] = API_NUM_PARALLEL_REQUESTS
  local_strategy: ClassVar[TaskExecutionType] = 'threads'

  @override
  def setup(self) -> None:
    api_key = env('PALM_API_KEY')
    gcp_project_id = env('GCP_PROJECT_ID')
    if not api_key and not gcp_project_id:
      raise ValueError(
        'Neither `PALM_API_KEY` or `GCP_PROJECT_ID` environment variables'
        ' are set, please provide one.'
      )
    if api_key:
      try:
        import google.generativeai as palm

        palm.configure(api_key=api_key)
        self._connector = 'api'
        self._num_parallel_requests = API_NUM_PARALLEL_REQUESTS
        self._model = palm.generate_embeddings
      except ImportError:
        raise ImportError(
          'Could not import the "google.generativeai" python package. '
          'Please install it with `pip install google-generativeai`.'
        )
    if gcp_project_id:
      try:
        from google.cloud import aiplatform
        from vertexai.language_models import TextEmbeddingModel

        aiplatform.init(project=gcp_project_id)
        self._connector = 'vertex'
        self._model = TextEmbeddingModel.from_pretrained(GCP_EMBEDDING_MODEL)
        self._num_parallel_requests = GCP_NUM_PARALLEL_REQUESTS
      except ImportError:
        raise ImportError(
          'Could not import the "vertex" python package. '
          'Please install it with `pip install google-cloud-aiplatform`.'
        )

  @override
  def compute(self, docs: list[str]) -> list[Optional[Item]]:
    """Compute embeddings for the given documents."""
    if self._connector == 'api':

      @retry(wait=wait_random_exponential(min=1, max=20), stop=stop_after_attempt(10))
      def embed_fn(texts: list[str]) -> list[np.ndarray]:
        assert len(texts) == 1, 'PaLM API only supports batch size 1.'
        response = self._model(model=API_EMBEDDING_MODEL, text=texts[0])
        return [np.array(response['embedding'], dtype=np.float32)]

    else:
      assert self._connector == 'vertex'

      @retry(wait=wait_fixed(5), stop=stop_after_attempt(15))
      def embed_fn(texts: list[str]) -> list[np.ndarray]:
        assert len(texts) <= 5, 'Vertex currently support a batch size of 5.'
        response = self._model.get_embeddings([texts[0]])[0].values
        return [np.array(response, dtype=np.float32)]

    return chunked_compute_embedding(
      embed_fn, docs, self.local_batch_size, chunker=clustering_spacy_chunker
    )



"""Test the semantic search signal."""

from typing import cast

from pytest import approx

from ..schema import FLOAT32, INT32, Field
from .text_statistics import (
  FRAC_NON_ASCII,
  NUM_CHARS,
  READABILITY,
  TYPE_TOKEN_RATIO,
  TextStatisticsSignal,
)


def test_text_statistics_fields() -> None:
  signal = TextStatisticsSignal()
  signal.setup()
  assert signal.fields() == Field(
    fields={
      NUM_CHARS: Field(dtype=INT32),
      READABILITY: Field(dtype=FLOAT32),
      TYPE_TOKEN_RATIO: Field(dtype=FLOAT32),
      FRAC_NON_ASCII: Field(
        dtype=FLOAT32,
        bins=[('Low', None, 0.15), ('Medium', 0.15, 0.3), ('High', 0.3, None)],
      ),
    }
  )


def test_text_statistics_compute() -> None:
  signal = TextStatisticsSignal()
  signal.setup()

  scores = signal.compute(['hello', 'hello world'])
  assert list(scores) == [
    {NUM_CHARS: 5, READABILITY: approx(2.62), TYPE_TOKEN_RATIO: 0.0, FRAC_NON_ASCII: 0.0},
    {NUM_CHARS: 11, READABILITY: approx(3.12), TYPE_TOKEN_RATIO: 1.0, FRAC_NON_ASCII: 0.0},
  ]


def test_text_statistics_missing_value() -> None:
  signal = TextStatisticsSignal()
  signal.setup()

  scores = signal.compute(['hello', cast(str, None), 'everybody'])

  assert list(scores) == [
    {NUM_CHARS: 5, READABILITY: approx(2.62), TYPE_TOKEN_RATIO: 0.0, FRAC_NON_ASCII: 0.0},
    None,
    {NUM_CHARS: 9, READABILITY: approx(21.46), TYPE_TOKEN_RATIO: 0.0, FRAC_NON_ASCII: 0.0},
  ]



"""Test batch_utils.py."""
import itertools
from typing import Iterable

import numpy as np

from .batch_utils import (
  flat_batched_compute,
  flatten,
  flatten_iter,
  flatten_path_iter,
  unflatten,
  unflatten_iter,
)


def test_batched_compute() -> None:
  input = [[1], [2, 3], [4, 5]]
  batch_size = 2  # Does not evenly split any input

  def f(inputs: Iterable[int]) -> list[int]:
    return [x * x for x in inputs]

  assert list(flat_batched_compute(input, f, batch_size)) == [[1], [4, 9], [16, 25]]


def test_batched_compute_np() -> None:
  input = [
    [np.array([1, 1])],
    [np.array([2, 2]), np.array([3, 3])],
    [np.array([4, 4]), np.array([5, 5])],
  ]
  batch_size = 2  # Does not evenly split any input

  def f(inputs: Iterable[np.ndarray]) -> Iterable[float]:
    return [x[0] * x[0] for x in inputs]

  assert list(flat_batched_compute(input, f, batch_size)) == [[1], [4, 9], [16, 25]]


def test_flatten() -> None:
  a = [[1, 2], [3], [4, 5, 5]]
  result = list(flatten(a))
  assert result == [1, 2, 3, 4, 5, 5]


def test_unflatten() -> None:
  a = [[1, 2], [3], [4, 5, 5]]
  flat_a = list(flatten(a))
  result = list(unflatten(flat_a, a))
  assert result == [[1, 2], [3], [4, 5, 5]]


def test_deep_flatten() -> None:
  a = [[1, 2], [[3]], [4, 5, [5]]]
  result = list(flatten_iter(a))
  assert result == [1, 2, 3, 4, 5, 5]


def test_deep_flatten_np() -> None:
  input = [[np.array([1, 1])], [np.array([2, 2]), np.array([3, 3])]]
  result = list(flatten_iter(input))

  assert len(result) == 3
  np.testing.assert_array_equal(result[0], np.array([1, 1]))
  np.testing.assert_array_equal(result[1], np.array([2, 2]))
  np.testing.assert_array_equal(result[2], np.array([3, 3]))


def test_deep_unflatten() -> None:
  a = [[1, 2], [[3]], [4, 5, 5]]
  flat_a = list(flatten_iter(a))
  flat_f_a = [a * 2 for a in flat_a]
  result = list(unflatten_iter(flat_f_a, a))
  assert result == [[2, 4], [[6]], [8, 10, 10]]


def test_deep_unflatten_generator() -> None:
  a = ([1, 2], [[3]], [4, 5, 5])
  a_0, a_1 = itertools.tee(a, 2)
  flat_a = list(flatten_iter(a_0))
  flat_f_a = [a * 2 for a in flat_a]
  result = list(unflatten_iter(flat_f_a, a_1))
  assert result == [[2, 4], [[6]], [8, 10, 10]]


def test_deep_unflatten_primitive() -> None:
  original = 'hello'
  result = next(iter(unflatten_iter(['hello'], original)))
  assert result == 'hello'


def test_deep_unflatten_primitive_list() -> None:
  original = ['hello', 'world']
  result = list(unflatten_iter(['hello', 'world'], original))
  assert result == ['hello', 'world']


def test_deep_unflatten_np() -> None:
  input = [[np.array([1, 1])], [np.array([2, 2]), np.array([3, 3])]]
  result = list(unflatten_iter(flatten_iter(input), input))

  assert len(result) == 2
  np.testing.assert_array_equal(result[0], [np.array([1, 1])])
  np.testing.assert_array_equal(result[1], [np.array([2, 2]), np.array([3, 3])])


def test_flatten_path_iter() -> None:
  input = [
    {'names': [{'first': 'John', 'last': 'Doe'}, {'first': 'Jane', 'last': 'Doe'}]},
    {'names': [{'first': 'Blake', 'last': 'Smith'}, {'first': 'Rob', 'last': 'Smith'}]},
  ]
  result = list(flatten_path_iter(input, path=('*', 'names', '*', 'first')))
  assert result == ['John', 'Jane', 'Blake', 'Rob']

  result = list(flatten_path_iter(input, path=('*', 'names', '*')))
  assert result == [
    {'first': 'John', 'last': 'Doe'},
    {'first': 'Jane', 'last': 'Doe'},
    {'first': 'Blake', 'last': 'Smith'},
    {'first': 'Rob', 'last': 'Smith'},
  ]
  result = list(flatten_path_iter(input, path=('*', 'names')))
  assert result == [
    [{'first': 'John', 'last': 'Doe'}, {'first': 'Jane', 'last': 'Doe'}],
    [{'first': 'Blake', 'last': 'Smith'}, {'first': 'Rob', 'last': 'Smith'}],
  ]

  result = list(flatten_path_iter(input, path=tuple('*')))
  assert result == [
    {'names': [{'first': 'John', 'last': 'Doe'}, {'first': 'Jane', 'last': 'Doe'}]},
    {'names': [{'first': 'Blake', 'last': 'Smith'}, {'first': 'Rob', 'last': 'Smith'}]},
  ]



"""Fixtures for dataset tests."""
import os
import pathlib
from typing import Generator, Optional, Type

import freezegun.config
import pytest
from pytest_mock import MockerFixture

from .data.dataset import Dataset
from .data.dataset_duckdb import DatasetDuckDB
from .data.dataset_test_utils import TestProcessLogger, make_dataset
from .db_manager import set_default_dataset_cls
from .schema import Item, Schema

# Fixes https://github.com/huggingface/transformers/issues/24545
freezegun.config.configure(extend_ignore_list=['transformers', 'torch'])


@pytest.fixture(scope='function', params=[DatasetDuckDB])
def make_test_data(
  tmp_path: pathlib.Path, mocker: MockerFixture, request: pytest.FixtureRequest
) -> Generator:
  """A pytest fixture for creating temporary test datasets."""
  mocker.patch.dict(os.environ, {'LILAC_PROJECT_DIR': str(tmp_path)})
  dataset_cls: Type[Dataset] = request.param
  set_default_dataset_cls(dataset_cls)

  def _make_test_data(items: list[Item], schema: Optional[Schema] = None) -> Dataset:
    return make_dataset(dataset_cls, tmp_path, items, schema)

  # Return the factory for datasets that test methods can use.
  yield _make_test_data


@pytest.fixture(scope='function')
def test_process_logger(tmp_path: pathlib.Path) -> Generator:
  """A pytest fixture for creating a logger that can be read between workers and the scheduler."""
  # Return the factory for datasets that test methods can use.
  yield TestProcessLogger(tmp_path)



"""Unit tests for dataset.cluster()."""
from typing import ClassVar, Iterable, Iterator

import pytest
from pytest_mock import MockerFixture

from ..embeddings.jina import JinaV2Small
from ..formats import ShareGPT
from ..schema import ClusterInfo, Item, field, schema
from ..signal import TextSignal, clear_signal_registry, register_signal
from ..source import clear_source_registry, register_source
from . import clustering
from .clustering import (
  CATEGORY_ID,
  CATEGORY_MEMBERSHIP_PROB,
  CATEGORY_TITLE,
  CLUSTER_ID,
  CLUSTER_MEMBERSHIP_PROB,
  CLUSTER_TITLE,
)
from .dataset import DatasetManifest, MetadataSearch
from .dataset_test_utils import (
  TEST_DATASET_NAME,
  TEST_NAMESPACE,
  TestDataMaker,
  TestSource,
  enriched_item,
)


class TestSignal(TextSignal):
  name: ClassVar[str] = 'test_signal'

  def compute(self, data: Iterator[str]) -> Iterator[int]:
    for text_content in data:
      yield len(text_content)


@pytest.fixture(scope='module', autouse=True)
def setup_teardown() -> Iterable[None]:
  # Setup.
  clear_signal_registry()
  clear_source_registry()
  register_source(TestSource)
  register_signal(JinaV2Small)
  register_signal(TestSignal)

  # Unit test runs.
  yield

  # Teardown.
  clear_signal_registry()
  clear_source_registry()


def test_simple_clusters(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  texts: list[str] = [
    'Can you summarize this article',
    'Can you rewrite this in a simpler way',
    'Can you provide a short summary of the following text',
    'Can you simplify this text',
  ]
  dataset = make_test_data([{'text': t} for t in texts])

  def topic_fn(docs: list[tuple[str, float]]) -> str:
    if 'summar' in docs[0][0]:
      return 'summarization'
    elif 'simpl' in docs[0][0]:
      return 'simplification'
    return 'other'

  mocker.patch.object(clustering, 'generate_category', return_value='MockCategory')
  dataset.cluster('text', min_cluster_size=2, topic_fn=topic_fn)

  rows = list(dataset.select_rows(['text', 'text__cluster'], combine_columns=True))
  assert rows == [
    {
      'text': 'Can you summarize this article',
      'text__cluster': {
        'cluster_id': 0,
        'cluster_membership_prob': 1.0,
        'cluster_title': 'summarization',
        'category_id': 0,
        'category_membership_prob': 1.0,
        'category_title': 'MockCategory',
      },
    },
    {
      'text': 'Can you rewrite this in a simpler way',
      'text__cluster': {
        'cluster_id': 1,
        'cluster_membership_prob': 1.0,
        'cluster_title': 'simplification',
        'category_id': 1,
        'category_membership_prob': 1.0,
        'category_title': 'MockCategory',
      },
    },
    {
      'text': 'Can you provide a short summary of the following text',
      'text__cluster': {
        'cluster_id': 0,
        'cluster_membership_prob': 1.0,
        'cluster_title': 'summarization',
        'category_id': 0,
        'category_membership_prob': 1.0,
        'category_title': 'MockCategory',
      },
    },
    {
      'text': 'Can you simplify this text',
      'text__cluster': {
        'cluster_id': 1,
        'cluster_membership_prob': 1.0,
        'cluster_title': 'simplification',
        'category_id': 1,
        'category_membership_prob': 1.0,
        'category_title': 'MockCategory',
      },
    },
  ]

  rows = list(
    dataset.select_rows(
      ['text__cluster'],
      searches=[
        MetadataSearch(path='text__cluster.cluster_title', op='equals', value='summarization')
      ],
      combine_columns=True,
    )
  )
  assert rows == [
    {
      'text__cluster': {
        'cluster_id': 0,
        'cluster_membership_prob': 1.0,
        'cluster_title': enriched_item(
          'summarization',
          {
            'metadata_search(op=equals,value=summarization)': True,
          },
        ),
        'category_id': 0,
        'category_membership_prob': 1.0,
        'category_title': 'MockCategory',
      },
    },
    {
      'text__cluster': {
        'cluster_id': 0,
        'cluster_membership_prob': 1.0,
        'cluster_title': enriched_item(
          'summarization',
          {
            'metadata_search(op=equals,value=summarization)': True,
          },
        ),
        'category_id': 0,
        'category_membership_prob': 1.0,
        'category_title': 'MockCategory',
      },
    },
  ]

  rows = list(
    dataset.select_rows(
      ['text__cluster'],
      searches=[
        MetadataSearch(path='text__cluster.category_title', op='equals', value='non_existent')
      ],
      combine_columns=True,
    )
  )
  assert rows == []

  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'text': 'string',
        'text__cluster': field(
          fields={
            CLUSTER_ID: field('int32', categorical=True),
            CLUSTER_MEMBERSHIP_PROB: 'float32',
            CLUSTER_TITLE: 'string',
            CATEGORY_ID: field('int32', categorical=True),
            CATEGORY_MEMBERSHIP_PROB: 'float32',
            CATEGORY_TITLE: 'string',
          },
          cluster=ClusterInfo(min_cluster_size=2, remote=False, input_path=('text',)),
        ),
      }
    ),
    num_items=4,
    source=TestSource(),
  )


def test_nested_clusters(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  texts: list[list[dict[str, str]]] = [
    [  # Cluster 1
      {'text': 'Can you summarize this article'},
      {'text': 'Can you provide a short summary of the following text'},
    ],
    [  # Cluster 2
      {'text': 'Can you rewrite this in a simpler way'},
      {'text': 'Can you simplify this text'},
    ],
    [  # Cluster 1
      {'text': 'Can you give a summary of the article'},
      {'text': 'Write a short overview of the following text'},
    ],
    [  # Cluster 2
      {'text': 'Can you write a simpler version'},
      {'text': 'Give me simplified version of this text'},
    ],
  ]
  mocker.patch.object(clustering, 'generate_category', return_value='MockCategory')
  dataset = make_test_data([{'texts': t} for t in texts])

  def topic_fn(docs: list[tuple[str, float]]) -> str:
    if 'summar' in docs[0][0]:
      return 'summarization'
    elif 'simpl' in docs[0][0]:
      return 'simplification'
    return 'other'

  dataset.cluster('texts.*.text', min_cluster_size=2, topic_fn=topic_fn)

  rows = list(dataset.select_rows(['texts_text__cluster'], combine_columns=True))
  assert rows == [
    {
      'texts_text__cluster': {
        'cluster_id': 0,
        'cluster_membership_prob': 1.0,
        'cluster_title': 'summarization',
        'category_id': 0,
        'category_membership_prob': 1.0,
        'category_title': 'MockCategory',
      },
    },
    {
      'texts_text__cluster': {
        'cluster_id': 1,
        'cluster_membership_prob': 1.0,
        'cluster_title': 'simplification',
        'category_id': 1,
        'category_membership_prob': 1.0,
        'category_title': 'MockCategory',
      },
    },
    {
      'texts_text__cluster': {
        'cluster_id': 0,
        'cluster_membership_prob': 1.0,
        'cluster_title': 'summarization',
        'category_id': 0,
        'category_membership_prob': 1.0,
        'category_title': 'MockCategory',
      },
    },
    {
      'texts_text__cluster': {
        'cluster_id': 1,
        'cluster_membership_prob': 1.0,
        'cluster_title': 'simplification',
        'category_id': 1,
        'category_membership_prob': 1.0,
        'category_title': 'MockCategory',
      },
    },
  ]


def test_path_ending_with_repeated(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  texts: list[list[str]] = [['hello', 'teacher'], ['professor'], ['hi']]
  dataset = make_test_data([{'texts': t} for t in texts])
  mocker.patch.object(clustering, 'generate_category', return_value='MockCategory')

  def topic_fn(docs: list[tuple[str, float]]) -> str:
    if 'hello' in docs[0][0]:
      return 'a_cluster'
    elif 'teacher' in docs[0][0]:
      return 'b_cluster'
    return 'other'

  dataset.cluster('texts.*', min_cluster_size=2, topic_fn=topic_fn)
  rows = list(dataset.select_rows(combine_columns=True))
  assert rows == [
    {
      'texts': ['hello', 'teacher'],
      'texts__cluster': {
        'cluster_id': -1,
        'cluster_membership_prob': 0.0,
        'cluster_title': None,
        'category_id': -1,
        'category_membership_prob': 0.0,
        'category_title': None,
      },
    },
    {
      'texts': ['professor'],
      'texts__cluster': {
        'cluster_id': -1,
        'cluster_membership_prob': 0.0,
        'cluster_title': None,
        'category_id': -1,
        'category_membership_prob': 0.0,
        'category_title': None,
      },
    },
    {
      'texts': ['hi'],
      'texts__cluster': {
        'cluster_id': -1,
        'cluster_membership_prob': 0.0,
        'cluster_title': None,
        'category_id': -1,
        'category_membership_prob': 0.0,
        'category_title': None,
      },
    },
  ]


def test_clusters_with_fn(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  texts: list[list[str]] = [
    ['Can you summarize this article'],
    ['Can you rewrite this in a simpler way'],
    ['Can you provide a short summary of the following text'],
    ['Can you simplify this text'],
  ]
  dataset = make_test_data([{'texts': t} for t in texts])
  mocker.patch.object(clustering, 'generate_category', return_value='MockCategory')

  def topic_fn(docs: list[tuple[str, float]]) -> str:
    if 'summar' in docs[0][0]:
      return 'summarization'
    elif 'simpl' in docs[0][0]:
      return 'simplification'
    return 'other'

  with pytest.raises(ValueError, match='output_path must be provided if input is a function'):
    dataset.cluster(lambda row: '\n'.join(row['texts']), min_cluster_size=2, topic_fn=topic_fn)

  dataset.cluster(
    lambda row: '\n'.join(row['texts']),
    output_path='cluster',
    min_cluster_size=2,
    topic_fn=topic_fn,
  )
  rows = list(dataset.select_rows(combine_columns=True))
  assert rows == [
    {
      'texts': ['Can you summarize this article'],
      'cluster': {
        'cluster_id': 0,
        'cluster_membership_prob': 1.0,
        'cluster_title': 'summarization',
        'category_id': 0,
        'category_membership_prob': 1.0,
        'category_title': 'MockCategory',
      },
    },
    {
      'texts': ['Can you rewrite this in a simpler way'],
      'cluster': {
        'cluster_id': 1,
        'cluster_membership_prob': 1.0,
        'cluster_title': 'simplification',
        'category_id': 1,
        'category_membership_prob': 1.0,
        'category_title': 'MockCategory',
      },
    },
    {
      'texts': ['Can you provide a short summary of the following text'],
      'cluster': {
        'cluster_id': 0,
        'cluster_membership_prob': 1.0,
        'cluster_title': 'summarization',
        'category_id': 0,
        'category_membership_prob': 1.0,
        'category_title': 'MockCategory',
      },
    },
    {
      'texts': ['Can you simplify this text'],
      'cluster': {
        'cluster_id': 1,
        'cluster_membership_prob': 1.0,
        'cluster_title': 'simplification',
        'category_id': 1,
        'category_membership_prob': 1.0,
        'category_title': 'MockCategory',
      },
    },
  ]


def test_clusters_with_fn_output_is_under_a_dict(
  make_test_data: TestDataMaker, mocker: MockerFixture
) -> None:
  texts: list[list[str]] = [
    ['Can you summarize this article'],
    ['Can you rewrite this in a simpler way'],
    ['Can you provide a short summary of the following text'],
    ['Can you simplify this text'],
  ]
  mocker.patch.object(clustering, 'generate_category', return_value='MockCategory')
  dataset = make_test_data([{'texts': t, 'info': {'dummy': True}} for t in texts])

  def topic_fn(docs: list[tuple[str, float]]) -> str:
    if 'summar' in docs[0][0]:
      return 'summarization'
    elif 'simpl' in docs[0][0]:
      return 'simplification'
    return 'other'

  dataset.cluster(
    lambda row: '\n'.join(row['texts']),
    output_path=('info', 'cluster'),
    min_cluster_size=2,
    topic_fn=topic_fn,
  )
  rows = list(dataset.select_rows(combine_columns=True))
  assert rows == [
    {
      'texts': ['Can you summarize this article'],
      'info': {
        'dummy': True,
        'cluster': {
          'cluster_id': 0,
          'cluster_membership_prob': 1.0,
          'cluster_title': 'summarization',
          'category_id': 0,
          'category_membership_prob': 1.0,
          'category_title': 'MockCategory',
        },
      },
    },
    {
      'texts': ['Can you rewrite this in a simpler way'],
      'info': {
        'dummy': True,
        'cluster': {
          'cluster_id': 1,
          'cluster_membership_prob': 1.0,
          'cluster_title': 'simplification',
          'category_id': 1,
          'category_membership_prob': 1.0,
          'category_title': 'MockCategory',
        },
      },
    },
    {
      'texts': ['Can you provide a short summary of the following text'],
      'info': {
        'dummy': True,
        'cluster': {
          'cluster_id': 0,
          'cluster_membership_prob': 1.0,
          'cluster_title': 'summarization',
          'category_id': 0,
          'category_membership_prob': 1.0,
          'category_title': 'MockCategory',
        },
      },
    },
    {
      'texts': ['Can you simplify this text'],
      'info': {
        'dummy': True,
        'cluster': {
          'cluster_id': 1,
          'cluster_membership_prob': 1.0,
          'cluster_title': 'simplification',
          'category_id': 1,
          'category_membership_prob': 1.0,
          'category_title': 'MockCategory',
        },
      },
    },
  ]


def test_clusters_sharegpt(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  mocker.patch.object(clustering, 'generate_category', return_value='MockCategory')

  texts: list[Item] = [
    {
      'conversations': [
        {'from': 'human', 'value': 'hello'},
        {'from': 'gpt', 'value': 'i am a language model'},
      ]
    },
    {
      'conversations': [
        {'from': 'human', 'value': 'whats the time'},
        {'from': 'gpt', 'value': '1030'},
      ]
    },
    {
      'conversations': [
        {'from': 'human', 'value': 'hello how are you'},
        {'from': 'gpt', 'value': 'pretty good today'},
      ]
    },
    {
      'conversations': [
        {'from': 'human', 'value': 'whats the hour'},
        {'from': 'gpt', 'value': '10 is the hour'},
      ]
    },
  ]
  dataset = make_test_data(texts)

  topic_fn_calls: list[list[tuple[str, float]]] = []

  def topic_fn(docs: list[tuple[str, float]]) -> str:
    topic_fn_calls.append(docs)
    if 'hello' in docs[0][0]:
      return 'greeting'
    elif 'time' in docs[0][0] or 'hour' in docs[0][0]:
      return 'time'
    return 'other'

  dataset.cluster(
    ShareGPT.human,
    output_path='cluster',
    min_cluster_size=2,
    topic_fn=topic_fn,
  )

  # Sort because topics are shuffled.
  for topic_fn_call in topic_fn_calls:
    topic_fn_call.sort()

  # Make sure the topic function is only called for the human text.
  assert topic_fn_calls == [
    [('hello', 1.0), ('hello how are you', 1.0)],
    [('whats the hour', 1.0), ('whats the time', 1.0)],
  ]

  rows = list(dataset.select_rows(combine_columns=True))
  assert rows == [
    {
      'conversations': [
        {'from': 'human', 'value': 'hello'},
        {'from': 'gpt', 'value': 'i am a language model'},
      ],
      'cluster': {
        'cluster_id': 0,
        'cluster_membership_prob': 1.0,
        'cluster_title': 'greeting',
        'category_id': 0,
        'category_membership_prob': 1.0,
        'category_title': 'MockCategory',
      },
    },
    {
      'conversations': [
        {'from': 'human', 'value': 'whats the time'},
        {'from': 'gpt', 'value': '1030'},
      ],
      'cluster': {
        'cluster_id': 1,
        'cluster_membership_prob': 1.0,
        'cluster_title': 'time',
        'category_id': 1,
        'category_membership_prob': 1.0,
        'category_title': 'MockCategory',
      },
    },
    {
      'conversations': [
        {'from': 'human', 'value': 'hello how are you'},
        {'from': 'gpt', 'value': 'pretty good today'},
      ],
      'cluster': {
        'cluster_id': 0,
        'cluster_membership_prob': 1.0,
        'cluster_title': 'greeting',
        'category_id': 0,
        'category_membership_prob': 1.0,
        'category_title': 'MockCategory',
      },
    },
    {
      'conversations': [
        {'from': 'human', 'value': 'whats the hour'},
        {'from': 'gpt', 'value': '10 is the hour'},
      ],
      'cluster': {
        'cluster_id': 1,
        'cluster_membership_prob': 1.0,
        'cluster_title': 'time',
        'category_id': 1,
        'category_membership_prob': 1.0,
        'category_title': 'MockCategory',
      },
    },
  ]


def test_clusters_on_enriched_text(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  texts: list[str] = [
    'Can you summarize this article',
    'Can you rewrite this in a simpler way',
    'Can you provide a short summary of the following text',
    'Can you simplify this text',
  ]
  mocker.patch.object(clustering, 'generate_category', return_value='MockCategory')
  dataset = make_test_data([{'text': t} for t in texts])

  def topic_fn(docs: list[tuple[str, float]]) -> str:
    if 'summar' in docs[0][0]:
      return 'summarization'
    elif 'simpl' in docs[0][0]:
      return 'simplification'
    return 'other'

  signal = TestSignal()
  dataset.compute_signal(signal, 'text')
  dataset.cluster('text', min_cluster_size=2, topic_fn=topic_fn)

  rows = list(dataset.select_rows(['text', 'text__cluster'], combine_columns=True))
  assert rows == [
    {
      'text': enriched_item('Can you summarize this article', {'test_signal': 30}),
      'text__cluster': {
        'cluster_id': 0,
        'cluster_membership_prob': 1.0,
        'cluster_title': 'summarization',
        'category_id': 0,
        'category_membership_prob': 1.0,
        'category_title': 'MockCategory',
      },
    },
    {
      'text': enriched_item('Can you rewrite this in a simpler way', {'test_signal': 37}),
      'text__cluster': {
        'cluster_id': 1,
        'cluster_membership_prob': 1.0,
        'cluster_title': 'simplification',
        'category_id': 1,
        'category_membership_prob': 1.0,
        'category_title': 'MockCategory',
      },
    },
    {
      'text': enriched_item(
        'Can you provide a short summary of the following text', {'test_signal': 53}
      ),
      'text__cluster': {
        'cluster_id': 0,
        'cluster_membership_prob': 1.0,
        'cluster_title': 'summarization',
        'category_id': 0,
        'category_membership_prob': 1.0,
        'category_title': 'MockCategory',
      },
    },
    {
      'text': enriched_item('Can you simplify this text', {'test_signal': 26}),
      'text__cluster': {
        'cluster_id': 1,
        'cluster_membership_prob': 1.0,
        'cluster_title': 'simplification',
        'category_id': 1,
        'category_membership_prob': 1.0,
        'category_title': 'MockCategory',
      },
    },
  ]



"""Utilities for testing text splitters."""

from typing import Optional, Union

from ..schema import SPAN_KEY, TEXT_SPAN_END_FEATURE, TEXT_SPAN_START_FEATURE, Item, span
from .chunk_splitter import TextChunk


def spans_to_text(text: str, spans: Optional[list[Item]]) -> list[str]:
  """Convert text and a list of spans to a list of strings."""
  if not spans:
    return []
  return [
    text[text_span[SPAN_KEY][TEXT_SPAN_START_FEATURE] : text_span[SPAN_KEY][TEXT_SPAN_END_FEATURE]]
    for text_span in spans
  ]


def text_to_expected_spans(
  text: str, splits: Union[list[str], list[tuple[str, Item]]], allowable_overlap: int = 0
) -> list[Item]:
  """Convert text and a list of splits to a list of expected spans."""
  start_offset = 0
  expected_spans: list[Item] = []
  for split in splits:
    item: Item
    if isinstance(split, str):
      split, item = split, {}
    elif isinstance(split, tuple):
      split, item = split
    else:
      raise ValueError('Split should be a string or a tuple of (string, item dict).')
    start = text.find(split, start_offset)
    end = start + len(split)
    expected_spans.append(span(start=start, end=end, metadata=item))
    start_offset = end - allowable_overlap

  return expected_spans


def text_to_textchunk(text: str, splits: list[str], allowable_overlap: int = 0) -> list[TextChunk]:
  """Convert text and a list of splits to a list of TextChunks."""
  spans = text_to_expected_spans(text, splits, allowable_overlap=allowable_overlap)
  expected_textchunks = []
  for text_span in spans:
    start, end = (
      text_span[SPAN_KEY][TEXT_SPAN_START_FEATURE],
      text_span[SPAN_KEY][TEXT_SPAN_END_FEATURE],
    )
    expected_textchunks.append((text[start:end], (start, end)))
  return expected_textchunks


def clean_textchunks(chunks: list[TextChunk]) -> list[TextChunk]:
  """Strip whitespace from TextChunks."""

  def clean_chunk(chunk: TextChunk) -> TextChunk:
    text, (start, _) = chunk
    stripped_text = text.strip()
    new_start = text.find(stripped_text) + start
    new_end = new_start + len(stripped_text)
    return stripped_text, (new_start, new_end)

  return [clean_chunk(chunk) for chunk in chunks]



"""Implementation of splitting text that looks at characters.

Recursively tries to split by different characters to find one that works.

The implementation below is forked from the LangChain project with the MIT license below.
See `RecursiveCharacterTextSplitter` in
https://github.com/hwchase17/langchain/blob/master/langchain/text_splitter.py
"""

# The MIT License

# Copyright (c) Harrison Chase

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from typing import Callable, Iterable, Optional

from ..utils import log

TextChunk = tuple[str, tuple[int, int]]

DEFAULT_SEPARATORS = ['```', '\n\n', '\n', ' ', '']
CHUNK_SIZE = 400
CHUNK_OVERLAP = 50


def _sep_split(text: str, separator: str) -> list[TextChunk]:
  if separator == '':
    # We need to split by char.
    return [(letter, (i, i + 1)) for i, letter in enumerate(text)]

  offset = 0
  chunks: list[TextChunk] = []
  open_code_block = False
  end_index = text.find(separator, offset)

  while end_index >= 0:
    if separator == '```':
      # We want to keep the code block seperators as part of the text chunk.
      start = max(0, offset - len(separator))
      if open_code_block:
        end = end_index + len(separator)
        open_code_block = False
      else:
        end = end_index
        open_code_block = True
    else:
      start = offset
      end = end_index

    chunks.append((text[start:end], (start, end)))
    offset = end_index + len(separator)
    end_index = text.find(separator, offset)

  # Append the last chunk.
  chunks.append((text[offset:], (offset, len(text))))

  return chunks


def split_text(
  text: str,
  chunk_size: int = CHUNK_SIZE,
  chunk_overlap: int = CHUNK_OVERLAP,
  separators: list[str] = DEFAULT_SEPARATORS,
  length_function: Callable[[str], int] = len,
) -> list[TextChunk]:
  """Split incoming text and return chunks."""
  text = str(text)

  def _merge_splits(splits: Iterable[TextChunk], separator: str) -> list[TextChunk]:
    # We now want to combine these smaller pieces into medium size chunks to send to the LLM.
    separator_len = 0 if separator == '```' else length_function(separator)

    docs: list[TextChunk] = []
    current_doc: list[TextChunk] = []
    total = 0
    for chunk in splits:
      text_chunk, _ = chunk
      _len = length_function(text_chunk)
      if total + _len + (separator_len if len(current_doc) > 0 else 0) > chunk_size:
        if total > chunk_size:
          log(
            f'Created a chunk of size {total}, ' f'which is longer than the specified {chunk_size}'
          )
        if len(current_doc) > 0:
          doc = _join_chunks(current_doc, separator)
          if doc is not None:
            docs.append(doc)
          # Keep on popping if:
          # - we have a larger chunk than in the chunk overlap
          # - or if we still have any chunks and the length is long
          while total > chunk_overlap or (
            total + _len + (separator_len if len(current_doc) > 0 else 0) > chunk_size and total > 0
          ):
            total -= length_function(current_doc[0][0]) + (
              separator_len if len(current_doc) > 1 else 0
            )
            current_doc = current_doc[1:]
      current_doc.append(chunk)
      total += _len + (separator_len if len(current_doc) > 1 else 0)
    doc = _join_chunks(current_doc, separator)
    if doc is not None:
      docs.append(doc)
    return docs

  final_chunks: list[TextChunk] = []
  # Get appropriate separator to use
  separator = separators[-1]
  new_separators: list[str] = []
  for i, _s in enumerate(separators):
    if _s == '':
      separator = _s
      break
    if _s in text:
      separator = _s
      new_separators = separators[i + 1 :]
      break
  # Now that we have the separator, split the text.
  splits = _sep_split(text, separator)
  # Now go merging things, recursively splitting longer texts.
  good_splits: list[TextChunk] = []
  for chunk in splits:
    text_chunk, (start, _) = chunk
    if length_function(text_chunk) < chunk_size:
      good_splits.append(chunk)
    else:
      if good_splits:
        merged_text = _merge_splits(good_splits, separator)
        final_chunks.extend(merged_text)
        good_splits = []
      other_chunks = split_text(
        text_chunk, chunk_size, chunk_overlap, new_separators, length_function
      )
      # Adjust the offsets of the other chunks.
      other_chunks = [(t, (s + start, e + start)) for t, (s, e) in other_chunks]
      final_chunks.extend(other_chunks)
  if good_splits:
    merged_text = _merge_splits(good_splits, separator)
    final_chunks.extend(merged_text)
  return final_chunks


def _join_chunks(chunks: list[TextChunk], separator: str) -> Optional[TextChunk]:
  if separator == '```':
    # Code blocks already have the separator.
    separator = ''
  text = separator.join([text for text, _ in chunks])
  text = text.strip()
  if text == '':
    return None

  _, (first_span_start, _) = chunks[0]
  _, (_, last_span_end) = chunks[-1]
  return (text, (first_span_start, last_span_end))



"""ShareGPT format."""

from typing import ClassVar

from ..dataset_format import DatasetFormat
from ..schema import PATH_WILDCARD, PathTuple, Schema, schema


# https://github.com/imoneoi/openchat
class OpenChat(DatasetFormat):
  """OpenChat format."""

  name: ClassVar[str] = 'openchat'
  data_schema: Schema = schema(
    {
      'items': [
        {
          'role': 'string',
          'content': 'string',
        }
      ],
      'system': 'string',
    },
  )

  title_slots: list[tuple[PathTuple, PathTuple]] = [
    (('items', PATH_WILDCARD, 'content'), ('items', PATH_WILDCARD, 'role'))
  ]



"""Interface for implementing a source."""

from typing import Any, Callable, ClassVar, Iterable, Optional, Type, Union

import numpy as np
import pandas as pd
import pyarrow as pa
from fastapi import APIRouter
from pydantic import BaseModel, ConfigDict, SerializeAsAny, field_validator, model_serializer

from .schema import (
  Field,
  ImageInfo,
  Item,
  Schema,
  arrow_dtype_to_dtype,
  arrow_schema_to_schema,
  field,
)
from .tasks import TaskId


class SourceSchema(BaseModel):
  """The schema of a source."""

  fields: dict[str, Field]
  num_items: Optional[int] = None


class SourceProcessResult(BaseModel):
  """The result after processing all the shards of a source dataset."""

  filepaths: list[str]
  data_schema: Schema
  num_items: int
  images: Optional[list[ImageInfo]] = None


def _source_schema_extra(schema: dict[str, Any], source: Type['Source']) -> None:
  """Add the title to the schema from the display name and name.

  Pydantic defaults this to the class name.
  """
  signal_prop: dict[str, Any]
  if hasattr(source, 'name'):
    signal_prop = {'enum': [source.name]}
  else:
    signal_prop = {'type': 'string'}
  schema['properties'] = {'source_name': signal_prop, **schema['properties']}
  if 'required' not in schema:
    schema['required'] = []
  schema['required'].append('source_name')


class Source(BaseModel):
  """Interface for sources to implement. A source processes a set of shards and writes files."""

  # ClassVars do not get serialized with pydantic.
  name: ClassVar[str]
  router: ClassVar[Optional[APIRouter]] = None

  @model_serializer(mode='wrap')
  def serialize_model(self, serializer: Callable[..., dict[str, Any]]) -> dict[str, Any]:
    """Serialize the model to a dictionary."""
    res = serializer(self)
    res['source_name'] = self.name
    return res

  model_config = ConfigDict(json_schema_extra=_source_schema_extra)

  def source_schema(self) -> SourceSchema:
    """Return the source schema for this source.

    Returns:
      A SourceSchema with
        fields: mapping top-level columns to fields that describes the schema of the source.
        num_items: the number of items in the source, used for progress.

    """
    raise NotImplementedError

  def setup(self) -> None:
    """Prepare the source for processing.

    This allows the source to do setup outside the constructor, but before its processed. This
    avoids potentially expensive computation the pydantic model is deserialized.
    """
    pass

  def teardown(self) -> None:
    """Tears down the source after processing."""
    pass

  def yield_items(self) -> Iterable[Item]:
    """Process the source by yielding individual rows of the source data.

    You should only override one of `yield_items` or `load_to_parquet`.

    This method is easier to use, and simply requires you to return an iterator of Python dicts.
    Lilac will take your iterable of items and handle writing it to parquet. You will still have to
    override source_schema.
    """
    raise NotImplementedError

  def load_to_parquet(self, output_dir: str, task_id: Optional[TaskId]) -> 'SourceManifest':
    """Process the source by directly writing a parquet file.

    You should only override one of `yield_items` or `load_to_parquet`.

    This fast path exists for sources where we are able to avoid the overhead of creating python
    dicts for every row by using non-Python parquet writers like DuckDB.

    The output parquet files should have a {schema.ROWID} column defined. This ROWID should not be
    part of source_schema, however.

    Finally, self.source_schema().num_items is usually computed in setup(), but can be left as None
    for sources implementing load_to_parquet, since this count is only used to display a progress
    bar. load_to_parquet doesn't have progress bar support so the count is unnecessary. However, you
    must still keep track of how many items were processed in total, because fields like
    self.sample_size should reflect the actual size of the dataset if len(dataset) < sample_size.

    Args:
      output_dir: The directory to write the parquet files to.
      task_id: The TaskManager id for this task. This is used to update
        the progress of the task.

    Returns:
      A SourceManifest that describes schema and parquet file locations.
    """
    raise NotImplementedError


class NoSource(Source):
  """A dummy source that is used when no source is defined, for backwards compat."""

  name: ClassVar[str] = 'no_source'


class SourceManifest(BaseModel):
  """The raw dataset contents, converted to Parquet, prior to any Lilac-specific annotations."""

  model_config = ConfigDict(strict=False)

  # List of a parquet filepaths storing the data. The paths can be relative to `manifest.json`.
  files: list[str]
  # The data schema.
  data_schema: Schema
  source: SerializeAsAny[Source] = NoSource()

  # Image information for the dataset.
  images: Optional[list[ImageInfo]] = None

  @field_validator('source', mode='before')
  @classmethod
  def parse_source(cls, source: dict) -> Source:
    """Parse a source to its specific subclass instance."""
    return resolve_source(source)


def schema_from_df(df: pd.DataFrame, index_colname: str) -> SourceSchema:
  """Create a source schema from a dataframe."""
  index_np_dtype = df.index.dtype
  # String index dtypes are stored as objects.
  if index_np_dtype == np.dtype(object):
    index_np_dtype = np.dtype(str)
  index_dtype = arrow_dtype_to_dtype(pa.from_numpy_dtype(index_np_dtype))

  schema = arrow_schema_to_schema(pa.Schema.from_pandas(df, preserve_index=False))
  return SourceSchema(
    fields={**schema.fields, index_colname: field(dtype=index_dtype)}, num_items=len(df)
  )


SOURCE_REGISTRY: dict[str, Type[Source]] = {}


def register_source(source_cls: Type[Source]) -> None:
  """Register a source configuration globally."""
  if source_cls.name in SOURCE_REGISTRY:
    raise ValueError(f'Source "{source_cls.name}" has already been registered!')

  SOURCE_REGISTRY[source_cls.name] = source_cls


def get_source_cls(source_name: str) -> Optional[Type[Source]]:
  """Return a registered source given the name in the registry."""
  return SOURCE_REGISTRY.get(source_name)


def registered_sources() -> dict[str, Type[Source]]:
  """Return all registered sources."""
  return SOURCE_REGISTRY


def resolve_source(source: Union[dict, Source]) -> Source:
  """Resolve a generic source base class to a specific source class."""
  if isinstance(source, Source):
    # The source is already parsed.
    return source

  source_name = source.get('source_name')
  if not source_name:
    raise ValueError('"source_name" needs to be defined in the json dict.')

  source_cls = get_source_cls(source_name)
  if not source_cls:
    # Make a metaclass so we get a valid `Source` class.
    source_cls = type(f'Source_{source_name}', (Source,), {'name': source_name})
  return source_cls(**source)


def clear_source_registry() -> None:
  """Clear the source registry."""
  SOURCE_REGISTRY.clear()



"""Loaders for managing the datasets on the Lilac demo."""

import gc
import os
import pathlib
import shutil
from typing import Optional, Union

from .concepts.db_concept import DiskConceptDB, DiskConceptModelDB
from .config import Config, EmbeddingConfig, SignalConfig, read_config
from .data.dataset_duckdb import DatasetDuckDB
from .dataset_format import DatasetFormatInputSelector, get_dataset_format_cls
from .db_manager import get_dataset, list_datasets, remove_dataset_from_cache
from .env import get_project_dir
from .load_dataset import process_source
from .project import PROJECT_CONFIG_FILENAME
from .schema import ROWID, PathTuple
from .utils import DebugTimer, get_datasets_dir, log


def load(
  project_dir: Optional[Union[str, pathlib.Path]] = None,
  config: Optional[Union[str, pathlib.Path, Config]] = None,
  overwrite: bool = False,
) -> None:
  """Load a project from a project configuration.

  Args:
    project_dir: The path to the project directory for where to create the dataset. If not defined,
      uses the project directory from `LILAC_PROJECT_DIR` or [deprecated] `LILAC_DATA_PATH`. The
      project_dir can be set globally with `set_project_dir`.
    config: A Lilac config or the path to a json or yml file describing the configuration. The
      contents should be an instance of `lilac.Config` or `lilac.DatasetConfig`. When not defined,
      uses `LILAC_PROJECT_DIR`/lilac.yml.
    overwrite: When True, runs all data from scratch, overwriting existing data. When false, only
      load new datasets, embeddings, and signals.
    execution_type: The execution type for the task manager. Can be 'processes' or 'threads'.
  """
  project_dir = project_dir or get_project_dir()
  if not project_dir:
    raise ValueError(
      '`project_dir` must be defined. Please pass a `project_dir` or set it '
      'globally with `set_project_dir(path)`'
    )

  # Turn off debug logging.
  if 'DEBUG' in os.environ:
    del os.environ['DEBUG']

  if not isinstance(config, Config):
    config_path = config or os.path.join(project_dir, PROJECT_CONFIG_FILENAME)
    config = read_config(config_path)

  if overwrite:
    shutil.rmtree(get_datasets_dir(project_dir), ignore_errors=True)

  existing_datasets = [f'{d.namespace}/{d.dataset_name}' for d in list_datasets(project_dir)]

  log()
  log('*** Load datasets ***')
  if overwrite:
    datasets_to_load = config.datasets
  else:
    datasets_to_load = [
      d for d in config.datasets if f'{d.namespace}/{d.name}' not in existing_datasets
    ]
    skipped_datasets = [
      d for d in config.datasets if f'{d.namespace}/{d.name}' in existing_datasets
    ]
    log('Skipping loaded datasets:', ', '.join([d.name for d in skipped_datasets]))

  with DebugTimer(f'Loading datasets: {", ".join([d.name for d in datasets_to_load])}'):
    for d in datasets_to_load:
      shutil.rmtree(os.path.join(project_dir, d.name), ignore_errors=True)
      process_source(project_dir, d)

  log()
  total_num_rows = 0
  for d in datasets_to_load:
    dataset = DatasetDuckDB(d.namespace, d.name, project_dir=project_dir)
    num_rows = dataset.select_rows([ROWID], limit=1).total_num_rows
    log(f'{d.namespace}/{d.name} loaded with {num_rows:,} rows.')

    # Free up RAM.
    del dataset

    total_num_rows += num_rows

  log(f'Done loading {len(datasets_to_load)} datasets with {total_num_rows:,} rows.')

  log('*** Dataset settings ***')
  for d in config.datasets:
    if d.settings:
      dataset = DatasetDuckDB(d.namespace, d.name, project_dir=project_dir)
      dataset.update_settings(d.settings)
      del dataset

  log()
  log('*** Compute embeddings ***')
  with DebugTimer('Loading embeddings'):
    for d in config.datasets:
      dataset = DatasetDuckDB(d.namespace, d.name, project_dir=project_dir)
      manifest = dataset.manifest()
      # If embeddings are explicitly set, use only those.
      embeddings = d.embeddings or []
      # If embeddings are not explicitly set, use the media paths and preferred embedding from
      # settings.
      if not embeddings:
        if d.settings and d.settings.ui:
          for path in d.settings.ui.media_paths or []:
            if d.settings.preferred_embedding:
              embeddings.append(
                EmbeddingConfig(path=path, embedding=d.settings.preferred_embedding)
              )
      for e in embeddings:
        field = manifest.data_schema.get_field(e.path)
        embedding_field = (field.fields or {}).get(e.embedding)
        if embedding_field is None or overwrite:
          _compute_embedding(
            d.namespace,
            d.name,
            e,
            project_dir,
          )
        else:
          log(f'Embedding {e.embedding} already exists for {d.name}:{e.path}. Skipping.')

      del dataset
      gc.collect()

  log()
  log('*** Compute signals ***')
  with DebugTimer('Computing signals'):
    for d in config.datasets:
      dataset = DatasetDuckDB(d.namespace, d.name, project_dir=project_dir)
      manifest = dataset.manifest()

      # If signals are explicitly set, use only those.
      signals = d.signals or []
      # If signals are not explicitly set, use the media paths and config.signals.
      if not signals:
        if d.settings and d.settings.ui:
          for path in d.settings.ui.media_paths or []:
            for signal in config.signals or []:
              signals.append(SignalConfig(path=path, signal=signal))

      # Separate signals by path to avoid computing the same signal in parallel, which can cause
      # issues with taking too much RAM.
      path_signals: dict[PathTuple, list[SignalConfig]] = {}
      for s in signals:
        path_signals.setdefault(s.path, []).append(s)

      for path, signals in path_signals.items():
        for s in signals:
          field = manifest.data_schema.get_field(s.path)
          signal_field = (field.fields or {}).get(s.signal.key(is_computed_signal=True))
          if signal_field is None or overwrite:
            _compute_signal(
              d.namespace,
              d.name,
              s,
              project_dir,
              overwrite,
            )
          else:
            log(f'Signal {s.signal} already exists for {d.name}:{s.path}. Skipping.')

      del dataset
      gc.collect()

  log()
  log('*** Compute clusters ***')
  with DebugTimer('Computing clusters'):
    for c in config.clusters:
      dataset = DatasetDuckDB(c.dataset_namespace, c.dataset_name, project_dir=project_dir)
      manifest = dataset.manifest()
      schema = manifest.data_schema
      # Try to find the cluster if it is precomputed.
      for path, node in schema.all_fields:
        if node.cluster is not None and (
          node.cluster.input_path == c.input_path
          or (
            node.cluster.input_format_selector
            and c.input_selector
            and (
              node.cluster.input_format_selector.format == c.input_selector.format
              and node.cluster.input_format_selector.selector == c.input_selector.selector
            )
          )
        ):
          log('Cluster already computed:', c)
          break
      else:
        # No precomputed cluster found.
        log('Computing cluster:', c)

        cluster_input: Optional[Union[DatasetFormatInputSelector, PathTuple]] = c.input_path
        if c.input_selector:
          format_cls = get_dataset_format_cls(c.input_selector.format)
          if format_cls is None:
            raise ValueError(f'Unknown format: {c.input_selector.format}')

          format = format_cls()
          if format != manifest.dataset_format:
            raise ValueError(
              f'Cluster input format {c.input_selector.format} does not match '
              f'dataset format {manifest.dataset_format}'
            )

          cluster_input = format_cls.input_selectors[c.input_selector.selector]

        assert cluster_input is not None
        dataset.cluster(
          cluster_input,
          output_path=c.output_path,
          min_cluster_size=c.min_cluster_size,
          remote=c.remote,
        )

  log()
  log('*** Compute model caches ***')
  with DebugTimer('Computing model caches'):
    concept_db = DiskConceptDB(project_dir)
    concept_model_db = DiskConceptModelDB(concept_db, project_dir=project_dir)
    if config.concept_model_cache_embeddings:
      for concept_info in concept_db.list():
        for embedding in config.concept_model_cache_embeddings:
          log('Syncing concept model cache:', concept_info, embedding)
          concept_model_db.sync(
            concept_info.namespace, concept_info.name, embedding_name=embedding, create=True
          )

  log()
  log('Done!')


def _compute_signal(
  namespace: str,
  name: str,
  signal_config: SignalConfig,
  project_dir: Union[str, pathlib.Path],
  overwrite: bool = False,
) -> None:
  # Turn off debug logging.
  if 'DEBUG' in os.environ:
    del os.environ['DEBUG']

  dataset = get_dataset(namespace, name, project_dir)
  dataset.compute_signal(
    signal=signal_config.signal,
    path=signal_config.path,
    overwrite=overwrite,
  )

  # Free up RAM.
  remove_dataset_from_cache(namespace, name)
  del dataset
  gc.collect()


def _compute_embedding(
  namespace: str,
  name: str,
  embedding_config: EmbeddingConfig,
  project_dir: Union[str, pathlib.Path],
) -> None:
  # Turn off debug logging.
  if 'DEBUG' in os.environ:
    del os.environ['DEBUG']

  dataset = get_dataset(namespace, name, project_dir)
  dataset.compute_embedding(
    embedding=embedding_config.embedding,
    path=embedding_config.path,
    overwrite=True,
  )
  remove_dataset_from_cache(namespace, name)
  del dataset
  gc.collect()



"""Router for the dataset database."""
import os
from copy import copy
from typing import Annotated, Any, Literal, Optional, Sequence, Union, cast

from fastapi import APIRouter, HTTPException, Response
from fastapi.params import Depends
from fastapi.responses import FileResponse, ORJSONResponse
from pydantic import BaseModel, Field

from .auth import UserInfo, get_session_user, get_user_access
from .config import DatasetSettings
from .data.dataset import (
  BinaryOp,
  DatasetManifest,
  FeatureListValue,
  FeatureValue,
  GroupsSortBy,
  ListOp,
  PivotResult,
  Search,
  SelectGroupsResult,
  SelectRowsSchemaResult,
  SortOrder,
  StatsResult,
  StringOp,
  UnaryOp,
)
from .data.dataset import Column as DBColumn
from .data.dataset import Filter as PyFilter
from .db_manager import DatasetInfo, get_dataset, list_datasets
from .env import get_project_dir
from .router_utils import RouteErrorHandler
from .schema import Bin, Path, normalize_path
from .signal import Signal, TextEmbeddingSignal, TextSignal
from .signals.concept_labels import ConceptLabelsSignal
from .signals.concept_scorer import ConceptSignal
from .signals.semantic_similarity import SemanticSimilaritySignal
from .signals.substring_search import SubstringSignal
from .utils import to_yaml

router = APIRouter(route_class=RouteErrorHandler)


@router.get('/', response_model_exclude_none=True)
def get_datasets() -> list[DatasetInfo]:
  """List the datasets."""
  return list_datasets(get_project_dir())


class WebManifest(BaseModel):
  """Information about a dataset."""

  dataset_manifest: DatasetManifest


@router.get('/{namespace}/{dataset_name}')
def get_manifest(namespace: str, dataset_name: str) -> WebManifest:
  """Get the web manifest for the dataset."""
  dataset = get_dataset(namespace, dataset_name)
  res = WebManifest(dataset_manifest=dataset.manifest())
  # Avoids the error that Signal abstract class is not serializable.
  return cast(WebManifest, ORJSONResponse(res.model_dump(exclude_none=True)))


@router.delete('/{namespace}/{dataset_name}')
def delete_dataset(
  namespace: str, dataset_name: str, user: Annotated[Optional[UserInfo], Depends(get_session_user)]
) -> None:
  """Delete the dataset."""
  if not get_user_access(user).dataset.delete_dataset:
    raise HTTPException(401, 'User does not have access to delete this dataset.')

  dataset = get_dataset(namespace, dataset_name)
  dataset.delete()


class GetStatsOptions(BaseModel):
  """The request for the get stats endpoint."""

  leaf_path: Path


@router.post('/{namespace}/{dataset_name}/stats')
def get_stats(namespace: str, dataset_name: str, options: GetStatsOptions) -> StatsResult:
  """Get the stats for the dataset."""
  dataset = get_dataset(namespace, dataset_name)
  return dataset.stats(options.leaf_path)


class BinaryFilter(BaseModel):
  """A filter on a column."""

  path: Path
  op: BinaryOp
  value: FeatureValue


class StringFilter(BaseModel):
  """A filter on a column."""

  path: Path
  op: StringOp
  value: str


class UnaryFilter(BaseModel):
  """A filter on a column."""

  path: Path
  op: UnaryOp
  value: None = None


class ListFilter(BaseModel):
  """A filter on a column."""

  path: Path
  op: ListOp
  value: FeatureListValue


Filter = Union[BinaryFilter, StringFilter, UnaryFilter, ListFilter]

AllSignalTypes = Union[
  ConceptSignal,
  ConceptLabelsSignal,
  SubstringSignal,
  SemanticSimilaritySignal,
  TextEmbeddingSignal,
  TextSignal,
  Signal,
]


# We override the `Column` class so we can add explicitly all signal types for better OpenAPI spec.
class Column(DBColumn):
  """A column in the dataset."""

  signal_udf: Optional[AllSignalTypes] = None


SearchPy = Annotated[Search, Field(discriminator='type')]


class SelectRowsOptions(BaseModel):
  """The request for the select rows endpoint."""

  columns: Sequence[Union[Column, Path]] = []
  searches: Sequence[SearchPy] = []
  filters: Sequence[Filter] = []
  sort_by: Sequence[Path] = []
  sort_order: Optional[SortOrder] = SortOrder.DESC
  limit: Optional[int] = None
  offset: Optional[int] = None
  combine_columns: Optional[bool] = None
  include_deleted: bool = False


class SelectRowsSchemaOptions(BaseModel):
  """The request for the select rows schema endpoint."""

  columns: Sequence[Union[Path, Column]] = []
  searches: Sequence[SearchPy] = []
  sort_by: Sequence[Path] = []
  sort_order: Optional[SortOrder] = SortOrder.DESC
  combine_columns: Optional[bool] = None


class SelectRowsResponse(BaseModel):
  """The response for the select rows endpoint."""

  rows: list[dict]
  total_num_rows: int


def _exclude_none(obj: Any) -> Any:
  if isinstance(obj, dict):
    return {k: _exclude_none(v) for k, v in obj.items() if v is not None}
  if isinstance(obj, list):
    return [_exclude_none(v) for v in obj]
  return copy(obj)


@router.post('/{namespace}/{dataset_name}/select_rows')
def select_rows(
  namespace: str,
  dataset_name: str,
  options: SelectRowsOptions,
  user: Annotated[Optional[UserInfo], Depends(get_session_user)],
) -> SelectRowsResponse:
  """Select rows from the dataset database."""
  dataset = get_dataset(namespace, dataset_name)

  sanitized_filters = [
    PyFilter(path=normalize_path(f.path), op=f.op, value=f.value) for f in (options.filters or [])
  ]

  res = dataset.select_rows(
    columns=options.columns,
    searches=options.searches or [],
    filters=sanitized_filters,
    sort_by=options.sort_by,
    sort_order=options.sort_order,
    limit=options.limit,
    offset=options.offset,
    combine_columns=options.combine_columns or False,
    include_deleted=options.include_deleted,
    user=user,
  )

  rows = [_exclude_none(row) for row in res]
  return SelectRowsResponse(rows=rows, total_num_rows=res.total_num_rows)


@router.post('/{namespace}/{dataset_name}/select_rows_schema', response_model_exclude_none=True)
def select_rows_schema(
  namespace: str, dataset_name: str, options: SelectRowsSchemaOptions
) -> SelectRowsSchemaResult:
  """Select rows from the dataset database."""
  dataset = get_dataset(namespace, dataset_name)
  return dataset.select_rows_schema(
    columns=options.columns,
    searches=options.searches or [],
    sort_by=options.sort_by,
    sort_order=options.sort_order,
    combine_columns=options.combine_columns or False,
  )


class SelectGroupsOptions(BaseModel):
  """The request for the select groups endpoint."""

  leaf_path: Path
  filters: Sequence[Filter] = []
  sort_by: Optional[GroupsSortBy] = GroupsSortBy.COUNT
  sort_order: Optional[SortOrder] = SortOrder.DESC
  limit: Optional[int] = 100
  bins: Optional[list[Bin]] = None


@router.post('/{namespace}/{dataset_name}/select_groups')
def select_groups(
  namespace: str, dataset_name: str, options: SelectGroupsOptions
) -> SelectGroupsResult:
  """Select groups from the dataset database."""
  dataset = get_dataset(namespace, dataset_name)
  sanitized_filters = [
    PyFilter(path=normalize_path(f.path), op=f.op, value=f.value) for f in (options.filters or [])
  ]
  return dataset.select_groups(
    options.leaf_path,
    sanitized_filters,
    options.sort_by,
    options.sort_order,
    options.limit,
    options.bins,
  )


class PivotOptions(BaseModel):
  """The request for the pivot endpoint."""

  inner_path: Path
  outer_path: Path
  filters: Sequence[Filter] = []


@router.post('/{namespace}/{dataset_name}/pivot')
def pivot(namespace: str, dataset_name: str, options: PivotOptions) -> PivotResult:
  """REST endpoint for dataset.pivot."""
  dataset = get_dataset(namespace, dataset_name)
  sanitized_filters = [
    PyFilter(path=normalize_path(f.path), op=f.op, value=f.value) for f in (options.filters or [])
  ]
  return dataset.pivot(
    options.outer_path,
    options.inner_path,
    filters=sanitized_filters,
  )


@router.get('/{namespace}/{dataset_name}/media')
def get_media(namespace: str, dataset_name: str, item_id: str, leaf_path: str) -> Response:
  """Get the media for the dataset."""
  dataset = get_dataset(namespace, dataset_name)
  path = tuple(leaf_path.split('.'))
  result = dataset.media(item_id, path)
  # Return the response via HTTP.
  return Response(content=result.data)


class ExportOptions(BaseModel):
  """The request for the export dataset endpoint."""

  format: Literal['csv', 'json', 'parquet']
  filepath: str
  jsonl: Optional[bool] = False
  columns: Sequence[Path] = []
  include_labels: Sequence[str] = []
  exclude_labels: Sequence[str] = []
  # Note: "__deleted__" is "just" another label, and the UI
  # will default to adding the "__deleted__" label to the exclude_labels list. If the user wants
  # to include deleted items, they can remove the "__deleted__" label from the exclude_labels list.
  # This is in contrast to just about every other API endpoint, which does explicitly include a
  # flag for include_deleted.


@router.get('/serve_dataset')
def serve_dataset_file(filepath: str) -> FileResponse:
  """Serve the exported dataset file."""
  filepath = os.path.expanduser(filepath)
  return FileResponse(filepath)


@router.post('/{namespace}/{dataset_name}/export')
def export_dataset(namespace: str, dataset_name: str, options: ExportOptions) -> str:
  """Export the dataset to one of the supported file formats."""
  dataset = get_dataset(namespace, dataset_name)
  # Only create the directory if the filepath has a directory.
  dirname = os.path.dirname(options.filepath)
  if dirname:
    os.makedirs(dirname, exist_ok=True)

  if options.format == 'csv':
    dataset.to_csv(
      options.filepath, options.columns, [], options.include_labels, options.exclude_labels
    )
  elif options.format == 'json':
    dataset.to_json(
      options.filepath,
      options.jsonl or False,
      options.columns,
      [],
      options.include_labels,
      options.exclude_labels,
    )
  elif options.format == 'parquet':
    dataset.to_parquet(
      options.filepath, options.columns, [], options.include_labels, options.exclude_labels
    )
  else:
    raise ValueError(f'Unknown format: {options.format}')
  return options.filepath


@router.get('/{namespace}/{dataset_name}/config')
def get_config(
  namespace: str, dataset_name: str, format: Literal['yaml', 'json']
) -> Union[str, dict]:
  """Get the config for the dataset."""
  dataset = get_dataset(namespace, dataset_name)
  config_dict = dataset.config().model_dump(exclude_defaults=True, exclude_none=True)
  if format == 'yaml':
    return to_yaml(config_dict)
  return config_dict


@router.get('/{namespace}/{dataset_name}/settings')
def get_settings(namespace: str, dataset_name: str) -> DatasetSettings:
  """Get the settings for the dataset."""
  dataset = get_dataset(namespace, dataset_name)
  return dataset.settings()


@router.post('/{namespace}/{dataset_name}/settings', response_model_exclude_none=True)
def update_settings(
  namespace: str,
  dataset_name: str,
  settings: DatasetSettings,
  user: Annotated[Optional[UserInfo], Depends(get_session_user)],
) -> None:
  """Update settings for the dataset."""
  if not get_user_access(user).dataset.compute_signals:
    raise HTTPException(401, 'User does not have access to update the settings of this dataset.')

  dataset = get_dataset(namespace, dataset_name)
  dataset.update_settings(settings)
  return None


class AddLabelsOptions(BaseModel):
  """The request for the add labels endpoint."""

  label_name: str
  label_value: Optional[str] = 'true'
  row_ids: Sequence[str] = []
  searches: Sequence[SearchPy] = []
  filters: Sequence[Filter] = []


@router.post('/{namespace}/{dataset_name}/labels', response_model_exclude_none=True)
def add_labels(
  namespace: str,
  dataset_name: str,
  options: AddLabelsOptions,
  user: Annotated[Optional[UserInfo], Depends(get_session_user)],
) -> int:
  """Add a label to the dataset."""
  if not get_user_access(user).dataset.edit_labels:
    raise HTTPException(401, 'User does not have access to add labels to this dataset.')

  if (options.searches or options.filters) and not get_user_access(user).dataset.label_all:
    raise HTTPException(401, 'User does not have access to use the label-all feature.')

  sanitized_filters = [
    PyFilter(path=normalize_path(f.path), op=f.op, value=f.value) for f in (options.filters or [])
  ]

  dataset = get_dataset(namespace, dataset_name)
  if (
    not get_user_access(user).dataset.create_label_type
    and options.label_name not in dataset.get_label_names()
  ):
    raise HTTPException(401, 'User does not have access to create label types in this dataset.')

  return dataset.add_labels(
    name=options.label_name,
    value=options.label_value,
    row_ids=options.row_ids,
    searches=options.searches,
    filters=sanitized_filters,
  )


class RemoveLabelsOptions(BaseModel):
  """The request for the remove labels endpoint."""

  label_name: str
  row_ids: Sequence[str] = []
  searches: Sequence[SearchPy] = []
  filters: Sequence[Filter] = []


@router.delete('/{namespace}/{dataset_name}/labels', response_model_exclude_none=True)
def remove_labels(
  namespace: str,
  dataset_name: str,
  options: RemoveLabelsOptions,
  user: Annotated[Optional[UserInfo], Depends(get_session_user)],
) -> int:
  """Remove a label from the dataset."""
  if not get_user_access(user).dataset.edit_labels:
    raise HTTPException(401, 'User does not have access to remove labels from this dataset.')

  sanitized_filters = [
    PyFilter(path=normalize_path(f.path), op=f.op, value=f.value) for f in (options.filters or [])
  ]

  dataset = get_dataset(namespace, dataset_name)
  return dataset.remove_labels(
    name=options.label_name,
    row_ids=options.row_ids,
    searches=options.searches,
    filters=sanitized_filters,
  )


class DeleteRowsOptions(BaseModel):
  """The request for the delete rows endpoint."""

  row_ids: Sequence[str] = []
  searches: Sequence[SearchPy] = []
  filters: Sequence[Filter] = []


@router.delete('/{namespace}/{dataset_name}/delete_rows', response_model_exclude_none=True)
def delete_rows(
  namespace: str,
  dataset_name: str,
  options: DeleteRowsOptions,
  user: Annotated[Optional[UserInfo], Depends(get_session_user)],
) -> int:
  """Add a label to the dataset."""
  if not get_user_access(user).dataset.delete_rows:
    raise HTTPException(401, 'User does not have access to delete rows on this dataset.')

  sanitized_filters = [
    PyFilter(path=normalize_path(f.path), op=f.op, value=f.value) for f in (options.filters or [])
  ]

  dataset = get_dataset(namespace, dataset_name)
  return dataset.delete_rows(
    row_ids=options.row_ids,
    searches=options.searches,
    filters=sanitized_filters,
  )


class RestoreRowsOptions(BaseModel):
  """The request for the restore rows endpoint."""

  row_ids: Sequence[str] = []
  searches: Sequence[SearchPy] = []
  filters: Sequence[Filter] = []


@router.post('/{namespace}/{dataset_name}/restore_rows', response_model_exclude_none=True)
def restore_rows(
  namespace: str,
  dataset_name: str,
  options: RestoreRowsOptions,
  user: Annotated[Optional[UserInfo], Depends(get_session_user)],
) -> int:
  """Add a label to the dataset."""
  if not get_user_access(user).dataset.delete_rows:
    raise HTTPException(401, 'User does not have access to restore rows on this dataset.')

  sanitized_filters = [
    PyFilter(path=normalize_path(f.path), op=f.op, value=f.value) for f in (options.filters or [])
  ]

  dataset = get_dataset(namespace, dataset_name)
  return dataset.restore_rows(
    row_ids=options.row_ids,
    searches=options.searches,
    filters=sanitized_filters,
  )



"""An interface for generators."""

import abc
from typing import Optional

from pydantic import BaseModel


class TextGenerator(BaseModel):
  """An interface for text generators."""

  @abc.abstractmethod
  def generate(self, prompt: str) -> str:
    """Generate a completion for a prompt."""
    raise NotImplementedError('`generate` is not implemented.')

  def num_tokens(self, prompt: str) -> Optional[int]:
    """Count the number of tokens for a prompt."""
    pass



"""Tests for dataset.map()."""

import inspect
import re
import time
from typing import ClassVar, Iterable, Iterator, Optional
from unittest.mock import ANY

import pytest
from typing_extensions import override

from .. import tasks
from ..batch_utils import group_by_sorted_key_iter
from ..schema import (
  PATH_WILDCARD,
  VALUE_KEY,
  Field,
  Item,
  MapInfo,
  RichData,
  field,
  schema,
  span,
)
from ..signal import TextSignal, clear_signal_registry, register_signal
from ..source import clear_source_registry, register_source
from ..test_utils import (
  TEST_TIME,
  allow_any_datetime,
)
from .dataset import DatasetManifest, Filter, SelectGroupsResult, StatsResult
from .dataset_test_utils import (
  TEST_DATASET_NAME,
  TEST_NAMESPACE,
  TestDataMaker,
  TestProcessLogger,
  TestSource,
  enriched_item,
)

TEST_EXECUTION_TYPES = ['threads', 'processes']


class TestFirstCharSignal(TextSignal):
  name: ClassVar[str] = 'test_signal'

  @override
  def fields(self) -> Field:
    return field(fields={'len': 'int32', 'firstchar': 'string'})

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    return ({'len': len(text_content), 'firstchar': text_content[0]} for text_content in data)


@pytest.fixture(scope='module', autouse=True)
def setup_teardown() -> Iterable[None]:
  allow_any_datetime(DatasetManifest)

  # Setup.
  clear_signal_registry()
  register_source(TestSource)
  register_signal(TestFirstCharSignal)

  # Unit test runs.
  yield
  # Teardown.
  clear_source_registry()
  clear_signal_registry()


@pytest.mark.parametrize('num_jobs', [-1, 1, 2])
@pytest.mark.parametrize('execution_type', TEST_EXECUTION_TYPES)
def test_map(
  num_jobs: int, execution_type: tasks.TaskExecutionType, make_test_data: TestDataMaker
) -> None:
  dataset = make_test_data([{'text': 'a sentence'}, {'text': 'b sentence'}])

  def _map_fn(item: Item) -> Item:
    return item['text'].upper()

  # Write the output to a new column.
  dataset.map(_map_fn, output_path='text_upper', num_jobs=num_jobs, execution_type=execution_type)

  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'text': 'string',
        'text_upper': field(
          dtype='string',
          map=MapInfo(
            fn_name='_map_fn',
            fn_source=inspect.getsource(_map_fn),
            date_created=TEST_TIME,
          ),
        ),
      }
    ),
    num_items=2,
    source=TestSource(),
  )

  rows = list(dataset.select_rows([PATH_WILDCARD]))
  assert rows == [
    {
      'text': 'a sentence',
      'text_upper': 'A SENTENCE',
    },
    {
      'text': 'b sentence',
      'text_upper': 'B SENTENCE',
    },
  ]


@pytest.mark.parametrize('execution_type', TEST_EXECUTION_TYPES)
def test_map_idle_worker(
  execution_type: tasks.TaskExecutionType, make_test_data: TestDataMaker
) -> None:
  dataset = make_test_data(
    [
      {'text': 'a sentence'},
      {'text': 'b sentence'},
      {'text': 'c sentence'},
    ]
  )

  def _map_fn(item: Item) -> Item:
    return item['text'].upper()

  dataset.map(_map_fn, output_path='text_upper', num_jobs=4, execution_type=execution_type)
  rows = list(dataset.select_rows([PATH_WILDCARD]))
  assert len(rows) == 3


@pytest.mark.parametrize('num_jobs', [-1, 1, 2])
@pytest.mark.parametrize('execution_type', TEST_EXECUTION_TYPES)
def test_map_signal(
  num_jobs: int, execution_type: tasks.TaskExecutionType, make_test_data: TestDataMaker
) -> None:
  dataset = make_test_data([{'text': 'a sentence'}, {'text': 'b sentence'}])

  signal = TestFirstCharSignal()
  dataset.compute_signal(signal, 'text')

  def _map_fn(item: Item) -> Item:
    return {'result': f'{item["text"]["test_signal"]["firstchar"]}_{len(item["text"][VALUE_KEY])}'}

  # Write the output to a new column.
  dataset.map(_map_fn, output_path='output_text', num_jobs=num_jobs, execution_type=execution_type)

  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'text': field(
          'string',
          fields={'test_signal': field(fields={'len': 'int32', 'firstchar': 'string'}, signal=ANY)},
        ),
        'output_text': field(
          fields={'result': 'string'},
          map=MapInfo(
            fn_name='_map_fn',
            fn_source=inspect.getsource(_map_fn),
            date_created=TEST_TIME,
          ),
        ),
      }
    ),
    num_items=2,
    source=TestSource(),
  )

  rows = list(dataset.select_rows([PATH_WILDCARD]))
  assert rows == [
    {
      'text': 'a sentence',
      'text.test_signal.firstchar': 'a',
      'text.test_signal.len': 10,
      'output_text.result': 'a_10',
    },
    {
      'text': 'b sentence',
      'text.test_signal.firstchar': 'b',
      'text.test_signal.len': 10,
      'output_text.result': 'b_10',
    },
  ]


@pytest.mark.parametrize('num_jobs', [-1, 1, 2])
@pytest.mark.parametrize('execution_type', TEST_EXECUTION_TYPES)
def test_map_input_path(
  num_jobs: int, execution_type: tasks.TaskExecutionType, make_test_data: TestDataMaker
) -> None:
  dataset = make_test_data(
    [
      {'id': 0, 'text': 'a'},
      {'id': 1, 'text': 'b'},
      {'id': 2, 'text': 'c'},
    ]
  )

  def _upper(row: Item) -> Item:
    return str(row).upper()

  # Write the output to a new column.
  dataset.map(
    _upper,
    input_path='text',
    output_path='text_upper',
    num_jobs=num_jobs,
    execution_type=execution_type,
  )

  # The schema should not reflect the output of the map as it didn't finish.
  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'text': 'string',
        'id': 'int32',
        'text_upper': field(
          dtype='string',
          map=MapInfo(
            fn_name='_upper',
            input_path=('text',),
            fn_source=inspect.getsource(_upper),
            date_created=TEST_TIME,
          ),
        ),
      }
    ),
    num_items=3,
    source=TestSource(),
  )
  # The rows should not reflect the output of the unfinished map.
  rows = list(dataset.select_rows([PATH_WILDCARD]))
  assert rows == [
    {'text': 'a', 'id': 0, 'text_upper': 'A'},
    {'text': 'b', 'id': 1, 'text_upper': 'B'},
    {'text': 'c', 'id': 2, 'text_upper': 'C'},
  ]


@pytest.mark.parametrize('num_jobs', [-1, 1, 2])
@pytest.mark.parametrize('execution_type', TEST_EXECUTION_TYPES)
def test_map_input_path_nested(
  num_jobs: int, execution_type: tasks.TaskExecutionType, make_test_data: TestDataMaker
) -> None:
  dataset = make_test_data(
    [
      {'id': 0, 'texts': [{'value': 'a'}]},
      {'id': 1, 'texts': [{'value': 'b'}, {'value': 'c'}]},
      {'id': 2, 'texts': [{'value': 'd'}, {'value': 'e'}, {'value': 'f'}]},
    ]
  )

  def _upper(row: Item) -> Item:
    return str(row).upper()

  dataset.map(
    _upper,
    input_path=('texts', PATH_WILDCARD, 'value'),
    output_path='texts_upper',
    num_jobs=num_jobs,
    execution_type=execution_type,
  )

  # The schema should not reflect the output of the map as it didn't finish.
  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'texts': [{'value': 'string'}],
        'id': 'int32',
        'texts_upper': field(
          fields=['string'],
          map=MapInfo(
            fn_name='_upper',
            input_path=('texts', PATH_WILDCARD, 'value'),
            fn_source=inspect.getsource(_upper),
            date_created=TEST_TIME,
          ),
        ),
      }
    ),
    num_items=3,
    source=TestSource(),
  )
  # The rows should not reflect the output of the unfinished map.
  rows = list(dataset.select_rows([PATH_WILDCARD]))
  assert rows == [
    {'id': 0, 'texts.*.value': ['a'], 'texts_upper': ['A']},
    {'id': 1, 'texts.*.value': ['b', 'c'], 'texts_upper': ['B', 'C']},
    {'id': 2, 'texts.*.value': ['d', 'e', 'f'], 'texts_upper': ['D', 'E', 'F']},
  ]


@pytest.mark.parametrize('num_jobs', [-1, 1, 2])
@pytest.mark.parametrize('execution_type', TEST_EXECUTION_TYPES)
def test_map_continuation(
  num_jobs: int,
  execution_type: tasks.TaskExecutionType,
  make_test_data: TestDataMaker,
  test_process_logger: TestProcessLogger,
) -> None:
  # Some flakiness in this test.
  # Say you have 2 workers. Worker 1 gets id 0. Worker 2 gets id 1, 2, 3, 4, 5....
  # For whatever reason worker 2 gets all the cpu time and gets to item N, which throws.
  # The pool is accumulating finished values from worker 2, while waiting on worker 1 because it has
  # in-order guarantees. But when worker 2 throws the exception, the whole job fails, without a
  # single item having been returned.
  # Threads seem to be the flakiest, even with very large num_items, because the GIL is
  # conservative about switching threads. The time.sleep() seems to give the GIL a hint to switch.
  num_items = 1000
  dataset = make_test_data([{'id': i} for i in range(num_items)])

  def _map_fn(item: Item, first_run: bool) -> Item:
    time.sleep(0.0001)
    if first_run and item['id'] == num_items - 1:
      raise ValueError('Throwing')

    return item['id']

  def _map_fn_1(item: Item) -> Item:
    return _map_fn(item, first_run=True)

  def _map_fn_2(item: Item) -> Item:
    test_process_logger.log_event(item['id'])
    return _map_fn(item, first_run=False)

  # Write the output to a new column.
  with pytest.raises(Exception):
    dataset.map(_map_fn_1, output_path='map_id', num_jobs=num_jobs, execution_type=execution_type)

  # The schema should not reflect the output of the map as it didn't finish.
  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema({'id': 'int32'}),
    num_items=num_items,
    source=TestSource(),
  )
  # The rows should not reflect the output of the unfinished map.
  rows = list(dataset.select_rows([PATH_WILDCARD]))
  assert rows == [{'id': i} for i in range(num_items)]

  # Hardcode this to thread-type, because the test_process_logger takes 2 seconds to serialize.
  # We're only really testing whether the continuation is reusing the saved results, anyway.
  dataset.map(_map_fn_2, output_path='map_id', num_jobs=1, execution_type='threads')

  assert 0 not in test_process_logger.get_logs()

  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'id': 'int32',
        'map_id': field(
          dtype='int64',
          map=MapInfo(
            fn_name='_map_fn_2', fn_source=inspect.getsource(_map_fn_2), date_created=TEST_TIME
          ),
        ),
      }
    ),
    num_items=num_items,
    source=TestSource(),
  )

  rows = list(dataset.select_rows([PATH_WILDCARD]))
  assert sorted(rows, key=lambda obj: obj['id']) == [
    {'id': i, 'map_id': i} for i in range(num_items)
  ]


@pytest.mark.parametrize('num_jobs', [-1, 1, 2])
@pytest.mark.parametrize('execution_type', TEST_EXECUTION_TYPES)
def test_map_continuation_overwrite(
  num_jobs: int,
  execution_type: tasks.TaskExecutionType,
  make_test_data: TestDataMaker,
  test_process_logger: TestProcessLogger,
) -> None:
  dataset = make_test_data(
    [
      {'id': 0, 'text': 'a sentence'},
      {'id': 1, 'text': 'b sentence'},
      {'id': 2, 'text': 'c sentence'},
    ]
  )

  def _map_fn(item: Item, first_run: bool) -> Item:
    test_process_logger.log_event(item['id'])

    if first_run and item['id'] == 1:
      raise ValueError('Throwing')

    return item['id']

  def _map_fn_1(item: Item) -> Item:
    return _map_fn(item, first_run=True)

  def _map_fn_2(item: Item) -> Item:
    return _map_fn(item, first_run=False)

  # Write the output to a new column.
  with pytest.raises(Exception):
    dataset.map(_map_fn_1, output_path='map_id', num_jobs=num_jobs, execution_type=execution_type)

  test_process_logger.clear_logs()

  dataset.map(
    _map_fn_2,
    output_path='map_id',
    overwrite=True,
    num_jobs=num_jobs,
    execution_type=execution_type,
  )

  # Map should be called for all ids.
  assert set(sorted(test_process_logger.get_logs())) == set([0, 1, 2])

  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'text': 'string',
        'id': 'int32',
        'map_id': field(
          dtype='int64',
          map=MapInfo(
            fn_name='_map_fn_2', fn_source=inspect.getsource(_map_fn_2), date_created=TEST_TIME
          ),
        ),
      }
    ),
    num_items=3,
    source=TestSource(),
  )

  rows = list(dataset.select_rows([PATH_WILDCARD]))
  assert rows == [
    {'text': 'a sentence', 'id': 0, 'map_id': 0},
    {'text': 'b sentence', 'id': 1, 'map_id': 1},
    {'text': 'c sentence', 'id': 2, 'map_id': 2},
  ]


@pytest.mark.parametrize('num_jobs', [-1, 1, 2])
@pytest.mark.parametrize('execution_type', TEST_EXECUTION_TYPES)
def test_map_limit(
  num_jobs: int, execution_type: tasks.TaskExecutionType, make_test_data: TestDataMaker
) -> None:
  dataset = make_test_data([{'text': letter} for letter in 'abcdefghijklmnopqrstuvwxyz'])

  def _map_fn(item: Item) -> Item:
    return item['text'].upper()

  # Write the output to a new column.
  dataset.map(
    _map_fn, output_path='text_upper', num_jobs=num_jobs, execution_type=execution_type, limit=10
  )

  rows = list(
    dataset.select_rows([PATH_WILDCARD], filters=[Filter(path=('text_upper',), op='exists')])
  )
  assert len(rows) == 10


@pytest.mark.parametrize('num_jobs', [-1, 1, 2])
@pytest.mark.parametrize('execution_type', TEST_EXECUTION_TYPES)
def test_map_filter(
  num_jobs: int, execution_type: tasks.TaskExecutionType, make_test_data: TestDataMaker
) -> None:
  dataset = make_test_data([{'text': letter} for letter in 'abcdefghijklmnopqrstuvwxyz'])

  def _map_fn(item: Item) -> Item:
    return item['text'].upper()

  # Write the output to a new column.
  dataset.map(
    _map_fn,
    output_path='text_upper',
    num_jobs=num_jobs,
    execution_type=execution_type,
    filters=[Filter(path=('text',), op='in', value=['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i'])],
  )

  rows = list(
    dataset.select_rows([PATH_WILDCARD], filters=[Filter(path=('text_upper',), op='exists')])
  )
  assert len(rows) == 9


@pytest.mark.parametrize('num_jobs', [-1, 1, 2])
@pytest.mark.parametrize('execution_type', TEST_EXECUTION_TYPES)
@pytest.mark.parametrize('batch_size', [-1, 2, 3])
def test_map_batch(
  num_jobs: int,
  execution_type: tasks.TaskExecutionType,
  batch_size: Optional[int],
  make_test_data: TestDataMaker,
) -> None:
  dataset = make_test_data([{'text': letter} for letter in 'abcd'])

  def _map_fn(items: list[Item]) -> list[Item]:
    return [item['text'].upper() for item in items]

  # Write the output to a new column.
  dataset.map(
    _map_fn,
    output_path='text_upper',
    batch_size=batch_size,
    num_jobs=num_jobs,
    execution_type=execution_type,
  )

  rows = list(
    dataset.select_rows([PATH_WILDCARD], filters=[Filter(path=('text_upper',), op='exists')])
  )
  assert rows == [
    {
      'text': 'a',
      'text_upper': 'A',
    },
    {
      'text': 'b',
      'text_upper': 'B',
    },
    {
      'text': 'c',
      'text_upper': 'C',
    },
    {
      'text': 'd',
      'text_upper': 'D',
    },
  ]


@pytest.mark.parametrize('num_jobs', [-1, 1, 2])
@pytest.mark.parametrize('execution_type', TEST_EXECUTION_TYPES)
def test_map_deleted_interaction(
  num_jobs: int, execution_type: tasks.TaskExecutionType, make_test_data: TestDataMaker
) -> None:
  dataset = make_test_data([{'text': letter} for letter in 'abcde'])

  def _map_fn(item: Item) -> Item:
    return item['text'].upper()

  dataset.delete_rows(row_ids=['1', '2', '3'])

  dataset.map(_map_fn, output_path='text_upper', num_jobs=num_jobs, execution_type=execution_type)

  rows = list(dataset.select_rows(['text_upper'], include_deleted=True))
  assert rows == [
    {'text_upper': 'D'},
    {'text_upper': 'E'},
    {'text_upper': None},
    {'text_upper': None},
    {'text_upper': None},
  ]

  dataset.map(
    _map_fn,
    output_path='text_upper',
    num_jobs=num_jobs,
    execution_type=execution_type,
    overwrite=True,
    include_deleted=True,
  )

  rows = list(dataset.select_rows(['text_upper'], include_deleted=True))
  assert rows == [
    {'text_upper': 'A'},
    {'text_upper': 'B'},
    {'text_upper': 'C'},
    {'text_upper': 'D'},
    {'text_upper': 'E'},
  ]


@pytest.mark.parametrize('num_jobs', [-1, 1, 2])
@pytest.mark.parametrize('execution_type', TEST_EXECUTION_TYPES)
def test_map_overwrite(
  num_jobs: int, execution_type: tasks.TaskExecutionType, make_test_data: TestDataMaker
) -> None:
  dataset = make_test_data([{'text': 'a'}, {'text': 'b'}])

  prefix = '0'

  def _map_fn(item: Item) -> Item:
    nonlocal prefix
    return prefix + item['text']

  # Write the output to a new column.
  dataset.map(_map_fn, output_path='map_text', num_jobs=num_jobs, execution_type=execution_type)

  rows = list(dataset.select_rows([PATH_WILDCARD]))
  assert rows == [{'text': 'a', 'map_text': '0a'}, {'text': 'b', 'map_text': '0b'}]

  with pytest.raises(ValueError, match=' which already exists in the dataset'):
    dataset.map(_map_fn, output_path='map_text', execution_type=execution_type)

  prefix = '1'
  # Overwrite the output
  dataset.map(_map_fn, output_path='map_text', overwrite=True, execution_type=execution_type)

  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'text': 'string',
        'map_text': field(
          dtype='string',
          map=MapInfo(
            fn_name='_map_fn', fn_source=inspect.getsource(_map_fn), date_created=TEST_TIME
          ),
        ),
      }
    ),
    num_items=2,
    source=TestSource(),
  )

  rows = list(dataset.select_rows([PATH_WILDCARD]))
  assert rows == [{'text': 'a', 'map_text': '1a'}, {'text': 'b', 'map_text': '1b'}]
  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'text': 'string',
        'map_text': field(
          dtype='string',
          map=MapInfo(
            fn_name='_map_fn', fn_source=inspect.getsource(_map_fn), date_created=TEST_TIME
          ),
        ),
      }
    ),
    num_items=2,
    source=TestSource(),
  )


def test_map_overwrite_where_input_is_output(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'a'}, {'text': 'b'}])

  def prefix_text(text: str) -> str:
    return '_' + text

  dataset.map(prefix_text, input_path='text', output_path='prefixed_text', overwrite=True)
  rows = list(dataset.select_rows(['prefixed_text']))
  assert rows == [{'prefixed_text': '_a'}, {'prefixed_text': '_b'}]

  # Overwrite prefixed_text by adding another prefix.
  dataset.map(prefix_text, input_path='prefixed_text', output_path='prefixed_text', overwrite=True)
  rows = list(dataset.select_rows(['prefixed_text']))
  assert rows == [{'prefixed_text': '__a'}, {'prefixed_text': '__b'}]

  def prefix_text_in_row(row: Item) -> str:
    return '_' + row['prefixed_text']

  # Overwrite prefixed_text by adding yet another prefix, this time iterating over the entire row.
  dataset.map(prefix_text_in_row, output_path='prefixed_text', overwrite=True)
  rows = list(dataset.select_rows(['prefixed_text']))
  assert rows == [{'prefixed_text': '___a'}, {'prefixed_text': '___b'}]

  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'text': 'string',
        'prefixed_text': field(
          dtype='string',
          map=MapInfo(
            fn_name='prefix_text_in_row',
            fn_source=inspect.getsource(prefix_text_in_row),
            date_created=TEST_TIME,
          ),
        ),
      }
    ),
    num_items=2,
    source=TestSource(),
  )


@pytest.mark.parametrize('num_jobs', [-1, 1, 2])
@pytest.mark.parametrize('execution_type', TEST_EXECUTION_TYPES)
def test_map_output_path_returns_iterable(
  num_jobs: int, execution_type: tasks.TaskExecutionType, make_test_data: TestDataMaker
) -> None:
  dataset = make_test_data([{'text': 'a sentence'}, {'text': 'b sentence'}])

  def _map_fn(item: Item) -> Item:
    return item['text'].upper()

  # Write the output to a new column.
  result_rows = dataset.map(
    _map_fn, output_path='text_upper', num_jobs=num_jobs, execution_type=execution_type
  )

  assert list(result_rows) == [
    'A SENTENCE',
    'B SENTENCE',
  ]

  rows = list(dataset.select_rows([PATH_WILDCARD]))
  assert rows == [
    {
      'text': 'a sentence',
      'text_upper': 'A SENTENCE',
    },
    {
      'text': 'b sentence',
      'text_upper': 'B SENTENCE',
    },
  ]


@pytest.mark.parametrize('num_jobs', [-1, 1, 2])
@pytest.mark.parametrize('execution_type', TEST_EXECUTION_TYPES)
def test_map_no_output_col(
  num_jobs: int, execution_type: tasks.TaskExecutionType, make_test_data: TestDataMaker
) -> None:
  dataset = make_test_data([{'text': 'a sentence'}, {'text': 'b sentence'}])

  def _map_fn(item: Item) -> Item:
    return item['text'].upper()

  map_output = dataset.map(_map_fn, num_jobs=num_jobs, execution_type=execution_type)

  assert list(map_output) == [
    'A SENTENCE',
    'B SENTENCE',
  ]
  # Make sure we can iterate again.
  assert list(map_output) == [
    'A SENTENCE',
    'B SENTENCE',
  ]

  # The manifest should contain no new columns.
  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema({'text': field('string')}),
    num_items=2,
    source=TestSource(),
  )

  # The new columns should not be saved anywhere.
  rows = list(dataset.select_rows([PATH_WILDCARD]))
  assert rows == [{'text': 'a sentence'}, {'text': 'b sentence'}]


def test_map_chained(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'a sentence'}, {'text': 'b sentence'}])

  def _split_fn(item: Item) -> Item:
    return item['text'].split(' ')

  dataset.map(_split_fn, output_path='splits')

  def _rearrange_fn(item: Item) -> Item:
    return item['splits'][1] + ' ' + item['splits'][0]

  dataset.map(_rearrange_fn, output_path='rearrange')

  rows = list(dataset.select_rows([PATH_WILDCARD]))
  assert rows == [
    {'text': 'a sentence', 'splits': ['a', 'sentence'], 'rearrange': 'sentence a'},
    {'text': 'b sentence', 'splits': ['b', 'sentence'], 'rearrange': 'sentence b'},
  ]

  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'text': field('string'),
        'splits': field(
          fields=['string'],
          map=MapInfo(
            fn_name='_split_fn', fn_source=inspect.getsource(_split_fn), date_created=TEST_TIME
          ),
        ),
        'rearrange': field(
          'string',
          map=MapInfo(
            fn_name='_rearrange_fn',
            fn_source=inspect.getsource(_rearrange_fn),
            date_created=TEST_TIME,
          ),
        ),
      }
    ),
    num_items=2,
    source=TestSource(),
  )


def test_map_over_enriched_item(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'a sentence'}, {'text': 'b sentence'}])

  signal = TestFirstCharSignal()
  dataset.compute_signal(signal, 'text')

  def _map_fn(item: Item) -> Item:
    return {'result': f'{item["text"]["test_signal"]["firstchar"]}_{len(item["text"][VALUE_KEY])}'}

  # Write the output to a new column.
  dataset.map(_map_fn, output_path='output_text')

  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'text': field(
          'string',
          fields={'test_signal': field(fields={'len': 'int32', 'firstchar': 'string'}, signal=ANY)},
        ),
        'output_text': field(
          fields={'result': 'string'},
          map=MapInfo(
            fn_name='_map_fn', fn_source=inspect.getsource(_map_fn), date_created=TEST_TIME
          ),
        ),
      }
    ),
    num_items=2,
    source=TestSource(),
  )

  rows = list(dataset.select_rows([PATH_WILDCARD]))
  assert rows == [
    {
      'text': 'a sentence',
      'text.test_signal.firstchar': 'a',
      'text.test_signal.len': 10,
      'output_text.result': 'a_10',
    },
    {
      'text': 'b sentence',
      'text.test_signal.firstchar': 'b',
      'text.test_signal.len': 10,
      'output_text.result': 'b_10',
    },
  ]


def test_map_on_double_nested_input(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [
      {'texts': [['a', 'b'], ['c'], ['dd', 'e']]},
      {'texts': [['ff', 'ggg']]},
    ]
  )

  dataset.map(lambda x: len(x), 'texts', output_path='texts_len')
  dataset.map(lambda x: len(x), 'texts.*', output_path='texts[i]_len')
  dataset.map(lambda x: len(x), 'texts.*.*', output_path='texts[i][j]_len')

  rows = list(dataset.select_rows())
  assert rows == [
    {
      'texts': [['a', 'b'], ['c'], ['dd', 'e']],
      'texts_len': 3,
      'texts[i]_len': [2, 1, 2],
      'texts[i][j]_len': [[1, 1], [1], [2, 1]],
    },
    {
      'texts': [['ff', 'ggg']],
      'texts_len': 1,
      'texts[i]_len': [2],
      'texts[i][j]_len': [[2, 3]],
    },
  ]


def test_map_select_subfields_of_repeated_dicts(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [
      {
        'people': [
          {'name': 'A', 'phones': [1, 2, 3], 'address': 'a'},
          {'name': 'BB', 'phones': [4], 'address': 'b'},
          {'name': 'C', 'phones': [5], 'address': 'c'},
        ]
      },
      {
        'people': [
          {'name': 'D', 'phones': [6], 'address': 'd'},
          {'name': 'EEE', 'phones': [7, 8], 'address': 'e'},
        ]
      },
    ]
  )

  dataset.map(lambda x: len(x), 'people', output_path='people_len')
  dataset.map(lambda x: x['name'], 'people.*', output_path='person_name')
  dataset.map(lambda x: len(x), 'people.*.name', output_path='len_name')
  dataset.map(lambda x: len(x), 'people.*.phones', output_path='len_phones')
  dataset.map(lambda x: x - 1, 'people.*.phones.*', output_path='phones_minus_one')

  # No input path.
  dataset.map(lambda x: 'people' in x and 'name' in x['people'][0], output_path='has_name')

  rows = list(
    dataset.select_rows(
      ['people_len', 'person_name', 'len_name', 'len_phones', 'phones_minus_one', 'has_name']
    )
  )
  assert rows == [
    {
      'people_len': 3,
      'person_name': ['A', 'BB', 'C'],
      'len_name': [1, 2, 1],
      'len_phones': [3, 1, 1],
      'phones_minus_one': [[0, 1, 2], [3], [4]],
      'has_name': True,
    },
    {
      'people_len': 2,
      'person_name': ['D', 'EEE'],
      'len_name': [1, 3],
      'len_phones': [1, 2],
      'phones_minus_one': [[5], [6, 7]],
      'has_name': True,
    },
  ]


def test_map_nests_under_repeated(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [
      {'people': [{'name': 'A'}, {'name': 'BB'}, {'name': 'C'}]},
      {'people': [{'name': 'D'}, {'name': 'EEE'}]},
    ]
  )

  dataset.map(lambda x: len(x), 'people.*.name', output_path='people.*.len_name')

  rows = list(dataset.select_rows(combine_columns=True))
  assert rows == [
    {
      'people': [
        {'name': 'A', 'len_name': 1},
        {'name': 'BB', 'len_name': 2},
        {'name': 'C', 'len_name': 1},
      ]
    },
    {'people': [{'name': 'D', 'len_name': 1}, {'name': 'EEE', 'len_name': 3}]},
  ]

  rows = list(dataset.select_rows())
  assert rows == [
    {'people.*.name': ['A', 'BB', 'C'], 'people.*.len_name': [1, 2, 1]},
    {'people.*.name': ['D', 'EEE'], 'people.*.len_name': [1, 3]},
  ]


def test_map_nests_under_field_of_repeated(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [
      {'people': [{'name': 'A', 'info': {'extra': 1}}, {'name': 'BB'}, {'name': 'C'}]},
      {'people': [{'name': 'D'}, {'name': 'EEE'}]},
    ]
  )

  dataset.map(lambda x: len(x), 'people.*.name', output_path='people.*.info.len_name')

  rows = list(dataset.select_rows(combine_columns=True))
  assert rows == [
    {
      'people': [
        {'name': 'A', 'info': {'extra': 1, 'len_name': 1}},
        {'name': 'BB', 'info': {'len_name': 2}},
        {'name': 'C', 'info': {'len_name': 1}},
      ]
    },
    {
      'people': [
        {'name': 'D', 'info': {'len_name': 1}},
        {'name': 'EEE', 'info': {'len_name': 3}},
      ]
    },
  ]


def test_map_nests_under_a_parent_of_input_path(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [
      {
        'people': [
          {'name': 'A', 'address': {'zip': 1}},
          {'name': 'B'},
          {'name': 'C', 'address': {'zip': 2}},
        ]
      },
      {
        'people': [
          {'name': 'D', 'address': {'zip': 3}},
          {'name': 'E', 'address': {}},
        ]
      },
    ]
  )

  def zip_code_verifier(zip_code: int) -> bool:
    return zip_code != 2

  dataset.map(zip_code_verifier, 'people.*.address.zip', output_path='people.*.verified')

  rows = list(dataset.select_rows(combine_columns=True))
  assert rows == [
    {
      'people': [
        {'name': 'A', 'address': {'zip': 1}, 'verified': True},
        {'name': 'B', 'address': None, 'verified': None},
        {'name': 'C', 'address': {'zip': 2}, 'verified': False},
      ]
    },
    {
      'people': [
        {'name': 'D', 'address': {'zip': 3}, 'verified': True},
        {'name': 'E', 'address': {'zip': None}, 'verified': None},
      ]
    },
  ]


def test_transform_with_sort_by(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [
      {'text': 'a', 'cluster_id': 2},
      {'text': 'b', 'cluster_id': 1},
      {'text': 'c', 'cluster_id': 1},
      {'text': 'd', 'cluster_id': 2},
      {'text': 'e', 'cluster_id': 2},
      {'text': 'f', 'cluster_id': 3},
    ]
  )

  def aggregate_cluster(items: Iterator[str]) -> Iterator[Item]:
    groups = group_by_sorted_key_iter(items, lambda x: x)
    for group in groups:
      for _ in group:
        yield len(group)

  dataset.transform(
    aggregate_cluster, input_path='cluster_id', output_path='cluster_size', sort_by='cluster_id'
  )

  rows = list(dataset.select_rows())
  assert rows == [
    {'text': 'a', 'cluster_id': 2, 'cluster_size': 3},
    {'text': 'b', 'cluster_id': 1, 'cluster_size': 2},
    {'text': 'c', 'cluster_id': 1, 'cluster_size': 2},
    {'text': 'd', 'cluster_id': 2, 'cluster_size': 3},
    {'text': 'e', 'cluster_id': 2, 'cluster_size': 3},
    {'text': 'f', 'cluster_id': 3, 'cluster_size': 1},
  ]


def test_signal_on_map_output(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'abcd'}, {'text': 'efghi'}])

  def _map_fn(item: Item) -> Item:
    return item['text'] + ' ' + item['text']

  dataset.map(_map_fn, output_path='double')

  # Compute a signal on the map output.
  signal = TestFirstCharSignal()
  dataset.compute_signal(signal, 'double')

  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'text': 'string',
        'double': field(
          'string',
          fields={'test_signal': field(fields={'len': 'int32', 'firstchar': 'string'}, signal=ANY)},
          map=MapInfo(
            fn_name='_map_fn', fn_source=inspect.getsource(_map_fn), date_created=TEST_TIME
          ),
        ),
      }
    ),
    num_items=2,
    source=TestSource(),
  )

  assert list(dataset.select_rows([PATH_WILDCARD])) == [
    {
      'text': 'abcd',
      'double': 'abcd abcd',
      'double.test_signal.firstchar': 'a',
      'double.test_signal.len': 9,
    },
    {
      'text': 'efghi',
      'double': 'efghi efghi',
      'double.test_signal.firstchar': 'e',
      'double.test_signal.len': 11,
    },
  ]

  # Select a struct root.
  assert list(dataset.select_rows([('double', 'test_signal')])) == [
    {
      'double.test_signal': {'firstchar': 'a', 'len': 9},
    },
    {
      'double.test_signal': {'firstchar': 'e', 'len': 11},
    },
  ]

  # combine_columns=True.
  rows = list(dataset.select_rows([PATH_WILDCARD], combine_columns=True))
  assert rows == [
    {
      'text': 'abcd',
      'double': enriched_item('abcd abcd', {signal.key(): {'firstchar': 'a', 'len': 9}}),
    },
    {
      'text': 'efghi',
      'double': enriched_item('efghi efghi', {signal.key(): {'firstchar': 'e', 'len': 11}}),
    },
  ]


def test_map_nested_output_path(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'parent': {'text': 'a'}}, {'parent': {'text': 'abc'}}])

  def _map_fn(item: Item) -> Item:
    return {'value': len(item['parent']['text'])}

  dataset.map(_map_fn, output_path='parent.text.len')

  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'parent': field(
          fields={
            'text': field(
              'string',
              fields={
                'len': field(
                  fields={'value': 'int64'},
                  map=MapInfo(
                    fn_name='_map_fn',
                    fn_source=inspect.getsource(_map_fn),
                    date_created=TEST_TIME,
                  ),
                )
              },
            )
          }
        ),
      }
    ),
    num_items=2,
    source=TestSource(),
  )

  rows = list(dataset.select_rows([PATH_WILDCARD], combine_columns=True))
  assert rows == [
    {
      'parent': {'text': enriched_item('a', {'len': {'value': 1}})},
    },
    {
      'parent': {'text': enriched_item('abc', {'len': {'value': 3}})},
    },
  ]


def test_map_span(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'ab'}, {'text': 'bc'}])

  # Return coordinates of 'b'.
  def _map_fn(item: Item) -> Item:
    return [
      span(m.start(), m.end(), {'len': m.end() - m.start()}) for m in re.finditer('b', item['text'])
    ]

  dataset.map(_map_fn, output_path='text.b_span')

  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'text': field(
          'string',
          fields={
            'b_span': field(
              fields=[field(dtype='string_span', fields={'len': 'int64'})],
              map=MapInfo(
                fn_name='_map_fn',
                fn_source=inspect.getsource(_map_fn),
                date_created=TEST_TIME,
              ),
            )
          },
        )
      }
    ),
    num_items=2,
    source=TestSource(),
  )

  rows = list(dataset.select_rows([PATH_WILDCARD], combine_columns=True))
  assert rows == [
    {
      'text': enriched_item('ab', {'b_span': [span(1, 2, {'len': 1})]}),
    },
    {
      'text': enriched_item('bc', {'b_span': [span(0, 1, {'len': 1})]}),
    },
  ]


def test_map_ergonomics(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'a sentence'}, {'text': 'b sentence'}])

  def _fn_kw(item: Item) -> Item:
    return item['text'] + ' map'

  def _fn(x: Item) -> Item:
    return x['text'] + ' map'

  # Write the output to a new column.
  dataset.map(_fn_kw, output_path='_fn_kw')
  dataset.map(_fn, output_path='_fn')

  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'text': 'string',
        '_fn_kw': field(
          dtype='string',
          map=MapInfo(
            fn_name='_fn_kw', fn_source=inspect.getsource(_fn_kw), date_created=TEST_TIME
          ),
        ),
        '_fn': field(
          dtype='string',
          map=MapInfo(fn_name='_fn', fn_source=inspect.getsource(_fn), date_created=TEST_TIME),
        ),
      }
    ),
    num_items=2,
    source=TestSource(),
  )

  rows = list(dataset.select_rows([PATH_WILDCARD]))
  assert rows == [
    {
      'text': 'a sentence',
      '_fn_kw': 'a sentence map',
      '_fn': 'a sentence map',
    },
    {
      'text': 'b sentence',
      '_fn_kw': 'b sentence map',
      '_fn': 'b sentence map',
    },
  ]


def test_map_ergonomics_invalid_args(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'a sentence'}, {'text': 'b sentence'}])

  def _map_noargs() -> None:
    pass

  with pytest.raises(TypeError, match=re.escape('takes 0 positional arguments but 1 was given')):
    dataset.map(_map_noargs, output_path='_map_noargs')

  def _map_toomany_args(row: Item, job_id: int, extra_arg: int) -> None:
    pass

  with pytest.raises(TypeError, match=re.escape('missing 2 required positional arguments')):
    dataset.map(_map_toomany_args, output_path='_map_toomany_args')


def test_map_nested_output_another_cardinality_fails(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [{'text': 'abcd', 'parent': ['a', 'b']}, {'text': 'efghi', 'parent': ['c', 'd']}]
  )

  def _map_fn(item: Item) -> Item:
    res = item['text'] + ' ' + item['text']
    return res

  with pytest.raises(
    AssertionError,
    match=re.escape(
      "`input_path` None and `output_path` ('parent', '*', 'output') have different cardinalities"
    ),
  ):
    dataset.map(_map_fn, output_path='parent.*.output')


def test_map_with_span_resolving(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'abcd'}, {'text': 'efghi'}])

  def skip_first_and_last_letter(item: str) -> Item:
    return span(1, len(item) - 1)

  dataset.map(skip_first_and_last_letter, input_path='text', output_path='skip')

  result = dataset.select_groups('skip')
  assert result == SelectGroupsResult(too_many_distinct=False, counts=[('bc', 1), ('fgh', 1)])

  stats = dataset.stats('skip')
  assert stats == StatsResult(
    path=('skip',), total_count=2, approx_count_distinct=2, avg_text_length=2.0
  )

  rows = dataset.select_rows()
  assert list(rows) == [
    {'text': 'abcd', 'skip': span(1, 3)},
    {'text': 'efghi', 'skip': span(1, 4)},
  ]


def test_transform(make_test_data: TestDataMaker) -> None:
  def text_len(items: Iterator[Item]) -> Iterator[Item]:
    for item in items:
      yield len(item['text'])

  dataset = make_test_data([{'text': 'abcd'}, {'text': 'efghi'}])
  dataset.transform(text_len, output_path='text_len')

  rows = dataset.select_rows()
  assert list(rows) == [{'text': 'abcd', 'text_len': 4}, {'text': 'efghi', 'text_len': 5}]


def test_transform_with_input_path(make_test_data: TestDataMaker) -> None:
  def text_len(texts: Iterator[Item]) -> Iterator[Item]:
    for text in texts:
      yield len(text)

  dataset = make_test_data([{'text': 'abcd'}, {'text': 'efghi'}])
  dataset.transform(text_len, input_path='text', output_path='text_len')

  rows = dataset.select_rows()
  assert list(rows) == [{'text': 'abcd', 'text_len': 4}, {'text': 'efghi', 'text_len': 5}]


def test_transform_size_mismatch(make_test_data: TestDataMaker) -> None:
  def text_len(texts: Iterator[Item]) -> Iterator[Item]:
    for i, text in enumerate(texts):
      # Skip the first item.
      if i > 0:
        yield len(text)

  dataset = make_test_data([{'text': 'abcd'}, {'text': 'efghi'}])
  with pytest.raises(Exception):
    dataset.transform(text_len, input_path='text', output_path='text_len')


def test_map_outputs_long_string_that_promotes_to_media(make_test_data: TestDataMaker) -> None:
  def long_str(text: str) -> str:
    return text * 100

  dataset = make_test_data([{'text': 'abcd'}, {'text': 'efghi'}])
  dataset.map(long_str, input_path='text', output_path='text_long')

  assert ('text_long',) in dataset.settings().ui.media_paths



"""Tests for item.py."""

import pyarrow as pa
import pytest

from .schema import (
  BINARY,
  BOOLEAN,
  FLOAT16,
  FLOAT32,
  FLOAT64,
  INT16,
  INT32,
  PATH_WILDCARD,
  SPAN_KEY,
  STRING,
  STRING_SPAN,
  TEXT_SPAN_END_FEATURE,
  TEXT_SPAN_START_FEATURE,
  VALUE_KEY,
  Field,
  Item,
  MapType,
  arrow_schema_to_schema,
  column_paths_match,
  field,
  schema,
  schema_to_arrow_schema,
)

NESTED_TEST_SCHEMA = schema(
  {
    'person': {
      'name': 'string',
      'last_name': 'string_span',
      # Contains a double nested array of primitives.
      'data': [['float32']],
      # Contains a value and children.
      'description': field(
        'string',
        fields={
          'toxicity': 'float32',
          'sentences': [field('string_span', fields={'len': 'int32'})],
        },
      ),
    },
    'addresses': [
      {
        'city': 'string',
        'zipcode': 'int16',
        'current': 'boolean',
        'locations': [{'latitude': 'float16', 'longitude': 'float64'}],
      }
    ],
    'blob': 'binary',
  }
)
NESTED_TEST_ITEM: Item = {
  'person': {'name': 'Test Name', 'last_name': (5, 9)},
  'addresses': [
    {
      'city': 'a',
      'zipcode': 1,
      'current': False,
      'locations': [{'latitude': 1.5, 'longitude': 3.8}, {'latitude': 2.9, 'longitude': 15.3}],
    },
    {
      'city': 'b',
      'zipcode': 2,
      'current': True,
      'locations': [{'latitude': 11.2, 'longitude': 20.1}, {'latitude': 30.1, 'longitude': 40.2}],
    },
  ],
}


def test_field_ctor_validation() -> None:
  with pytest.raises(
    ValueError, match='One of "fields", "repeated_field", or "dtype" should be defined'
  ):
    Field()

  with pytest.raises(ValueError, match='Both "fields" and "repeated_field" should not be defined'):
    Field(fields={'name': Field(dtype=STRING)}, repeated_field=Field(dtype=INT32))

  with pytest.raises(ValueError, match=f'{VALUE_KEY} is a reserved field name'):
    Field(fields={VALUE_KEY: Field(dtype=STRING)})


def test_schema_leafs() -> None:
  expected = {
    ('addresses', PATH_WILDCARD, 'city'): Field(dtype=STRING),
    ('addresses', PATH_WILDCARD, 'current'): Field(dtype=BOOLEAN),
    ('addresses', PATH_WILDCARD, 'locations', PATH_WILDCARD, 'latitude'): Field(dtype=FLOAT16),
    ('addresses', PATH_WILDCARD, 'locations', PATH_WILDCARD, 'longitude'): Field(dtype=FLOAT64),
    ('addresses', PATH_WILDCARD, 'zipcode'): Field(dtype=INT16),
    ('blob',): Field(dtype=BINARY),
    ('person', 'name'): Field(dtype=STRING),
    ('person', 'last_name'): Field(dtype=STRING_SPAN),
    ('person', 'data', PATH_WILDCARD, PATH_WILDCARD): Field(dtype=FLOAT32),
    ('person', 'description'): Field(
      dtype=STRING,
      fields={
        'toxicity': Field(dtype=FLOAT32),
        'sentences': Field(
          repeated_field=Field(dtype=STRING_SPAN, fields={'len': Field(dtype=INT32)})
        ),
      },
    ),
    ('person', 'description', 'toxicity'): Field(dtype=FLOAT32),
    ('person', 'description', 'sentences', PATH_WILDCARD): Field(
      fields={'len': Field(dtype=INT32)}, dtype=STRING_SPAN
    ),
    ('person', 'description', 'sentences', PATH_WILDCARD, 'len'): Field(dtype=INT32),
  }
  assert NESTED_TEST_SCHEMA.leafs == expected


def test_schema_to_arrow_schema() -> None:
  arrow_schema = schema_to_arrow_schema(NESTED_TEST_SCHEMA)

  assert arrow_schema == pa.schema(
    {
      'person': pa.struct(
        {
          'name': pa.string(),
          # The dtype for STRING_SPAN is implemented as a struct with a {start, end}.
          'last_name': pa.struct(
            {
              SPAN_KEY: pa.struct(
                {TEXT_SPAN_START_FEATURE: pa.int32(), TEXT_SPAN_END_FEATURE: pa.int32()}
              )
            }
          ),
          'data': pa.list_(pa.list_(pa.float32())),
          'description': pa.struct(
            {
              'toxicity': pa.float32(),
              'sentences': pa.list_(
                pa.struct(
                  {
                    'len': pa.int32(),
                    SPAN_KEY: pa.struct(
                      {TEXT_SPAN_START_FEATURE: pa.int32(), TEXT_SPAN_END_FEATURE: pa.int32()}
                    ),
                  }
                )
              ),
              VALUE_KEY: pa.string(),
            }
          ),
        }
      ),
      'addresses': pa.list_(
        pa.struct(
          {
            'city': pa.string(),
            'zipcode': pa.int16(),
            'current': pa.bool_(),
            'locations': pa.list_(pa.struct({'latitude': pa.float16(), 'longitude': pa.float64()})),
          }
        )
      ),
      'blob': pa.binary(),
    }
  )


def test_arrow_schema_to_schema() -> None:
  arrow_schema = pa.schema(
    {
      'person': pa.struct({'name': pa.string(), 'data': pa.list_(pa.list_(pa.float32()))}),
      'addresses': pa.list_(
        pa.struct(
          {
            'city': pa.string(),
            'zipcode': pa.int16(),
            'current': pa.bool_(),
            'locations': pa.list_(pa.struct({'latitude': pa.float16(), 'longitude': pa.float64()})),
          }
        )
      ),
      'blob': pa.binary(),
      'spans': pa.struct(
        {SPAN_KEY: pa.struct({'start': pa.int32(), 'end': pa.int32()}), 'metadata': pa.string()}
      ),
      'map': pa.map_(pa.string(), pa.list_(pa.int32())),
    }
  )
  expected_schema = schema(
    {
      'person': {'name': 'string', 'data': [['float32']]},
      'addresses': [
        {
          'city': 'string',
          'zipcode': 'int16',
          'current': 'boolean',
          'locations': [{'latitude': 'float16', 'longitude': 'float64'}],
        }
      ],
      'blob': 'binary',
      'spans': field('string_span', fields={'metadata': 'string'}),
      'map': field(
        dtype=MapType(key_type=STRING, value_field=Field(repeated_field=Field(dtype=INT32)))
      ),
    }
  )
  assert arrow_schema_to_schema(arrow_schema) == expected_schema


def test_simple_schema_str() -> None:
  assert str(schema({'person': 'string'})) == 'person: string'


def test_column_paths_match() -> None:
  assert column_paths_match(path_match=('person', 'name'), specific_path=('person', 'name')) is True
  assert (
    column_paths_match(path_match=('person', 'name'), specific_path=('person', 'not_name')) is False
  )

  # Wildcards work for structs.
  assert (
    column_paths_match(path_match=(PATH_WILDCARD, 'name'), specific_path=('person', 'name')) is True
  )
  assert (
    column_paths_match(path_match=(PATH_WILDCARD, 'name'), specific_path=('person', 'not_name'))
    is False
  )

  # Wildcards work for repeateds.
  assert (
    column_paths_match(
      path_match=('person', PATH_WILDCARD, 'name'), specific_path=('person', '0', 'name')
    )
    is True
  )
  assert (
    column_paths_match(
      path_match=('person', PATH_WILDCARD, 'name'), specific_path=('person', '0', 'not_name')
    )
    is False
  )

  # Sub-path matches always return False.
  assert column_paths_match(path_match=(PATH_WILDCARD,), specific_path=('person', 'name')) is False
  assert (
    column_paths_match(path_match=('person', PATH_WILDCARD), specific_path=('person', '0', 'name'))
    is False
  )


def test_nested_schema_str() -> None:
  assert (
    str(NESTED_TEST_SCHEMA)
    == """\
person:
  name: string
  last_name: string_span
  data: list( list( float32))
  description:
    toxicity: float32
    sentences: list(
      len: int32)
addresses: list(
  city: string
  zipcode: int16
  current: boolean
  locations: list(
    latitude: float16
    longitude: float64))
blob: binary\
"""
  )



from importlib import metadata

from .concepts import *  # noqa: F403
from .config import (
  Config,
  DatasetConfig,
  DatasetSettings,
  DatasetUISettings,
  EmbeddingConfig,
  SignalConfig,
)
from .data import *  # noqa: F403
from .data.dataset_duckdb import DatasetDuckDB
from .data.dataset_storage_utils import download, upload
from .db_manager import get_dataset, list_datasets, set_default_dataset_cls
from .deploy import deploy_config, deploy_project
from .embeddings import *  # noqa: F403
from .env import *  # noqa: F403
from .env import LilacEnvironment, get_project_dir, set_project_dir
from .formats import *  # noqa: F403
from .formats import OpenChat, ShareGPT
from .load import load
from .load_dataset import create_dataset, from_dicts, from_huggingface
from .project import init
from .rag import *  # noqa: F403
from .schema import *  # noqa: F403
from .schema import Field, SpanVector, span
from .server import start_server, stop_server
from .signals import *  # noqa: F403
from .source import Source
from .sources import *  # noqa: F403
from .splitters import *  # noqa: F403

try:
  __version__ = metadata.version('lilac')
except metadata.PackageNotFoundError:
  __version__ = ''

set_default_dataset_cls(DatasetDuckDB)

# Avoids polluting the results of dir(__package__).
del (
  metadata,
  set_default_dataset_cls,
  DatasetDuckDB,
)

__all__ = [
  'start_server',
  'stop_server',
  'create_dataset',
  'from_dicts',
  'from_huggingface',
  'get_dataset',
  'list_datasets',
  'init',
  'span',
  'load',
  'set_project_dir',
  'get_project_dir',
  'Config',
  'DatasetConfig',
  'EmbeddingConfig',
  'SignalConfig',
  'DatasetSettings',
  'DatasetUISettings',
  'LilacEnvironment',
  'Source',
  'Field',
  'deploy_project',
  'deploy_config',
  'SpanVector',
  'ShareGPT',
  'OpenChat',
  'download',
  'upload',
]



"""Test our public dataset/signals computation REST API."""

import os
from typing import ClassVar, Iterable, Iterator, Optional

from fastapi.testclient import TestClient
from pytest_mock import MockerFixture

from .data.dataset_test_utils import TEST_DATASET_NAME, TEST_NAMESPACE
from .router_dataset_signals import ComputeSignalOptions, DeleteSignalOptions
from .schema import Field, Item, RichData, field
from .server import app
from .signal import TextSignal

client = TestClient(app)


class LengthSignal(TextSignal):
  name: ClassVar[str] = 'length_signal'

  def fields(self) -> Field:
    return field('int32')

  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    for text_content in data:
      yield len(text_content) if text_content is not None else None


def test_compute_signal_auth(mocker: MockerFixture) -> None:
  mocker.patch.dict(os.environ, {'LILAC_AUTH_ENABLED': 'True'})

  url = f'/api/v1/datasets/{TEST_NAMESPACE}/{TEST_DATASET_NAME}/compute_signal'
  response = client.post(
    url, json=ComputeSignalOptions(signal=LengthSignal(), leaf_path=('people', 'name')).model_dump()
  )
  assert response.status_code == 401
  assert response.is_error is True
  assert 'User does not have access to compute signals over this dataset.' in response.text


def test_delete_signal_auth(mocker: MockerFixture) -> None:
  mocker.patch.dict(os.environ, {'LILAC_AUTH_ENABLED': 'True'})

  url = f'/api/v1/datasets/{TEST_NAMESPACE}/{TEST_DATASET_NAME}/delete_signal'
  response = client.request(
    'DELETE', url, json=DeleteSignalOptions(signal_path=('doesnt', 'matter')).model_dump()
  )
  assert response.status_code == 401
  assert response.is_error is True
  assert 'User does not have access to delete this signal.' in response.text



"""NumpyVectorStore class for storing vectors in numpy arrays."""

from typing import Iterable, Optional, cast

import numpy as np
import pandas as pd
from typing_extensions import override

from ..schema import VectorKey
from .vector_store import VectorStore

_EMBEDDINGS_SUFFIX = '.matrix.npy'
_LOOKUP_SUFFIX = '.lookup.pkl'


class NumpyVectorStore(VectorStore):
  """Stores vectors as in-memory np arrays."""

  name = 'numpy'

  def __init__(self) -> None:
    self._embeddings: Optional[np.ndarray] = None
    # Maps a `VectorKey` to a row index in `_embeddings`.
    self._key_to_index: Optional[pd.Series] = None

  @override
  def size(self) -> int:
    assert (
      self._embeddings is not None
    ), 'The vector store has no embeddings. Call load() or add() first.'
    return len(self._embeddings)

  @override
  def save(self, base_path: str) -> None:
    assert (
      self._embeddings is not None and self._key_to_index is not None
    ), 'The vector store has no embeddings. Call load() or add() first.'
    np.save(base_path + _EMBEDDINGS_SUFFIX, self._embeddings, allow_pickle=False)
    self._key_to_index.to_pickle(base_path + _LOOKUP_SUFFIX)

  @override
  def load(self, base_path: str) -> None:
    self._embeddings = np.load(base_path + _EMBEDDINGS_SUFFIX, allow_pickle=False)
    self._key_to_index = pd.read_pickle(base_path + _LOOKUP_SUFFIX)

  @override
  def add(self, keys: list[VectorKey], embeddings: np.ndarray) -> None:
    if len(keys) != embeddings.shape[0]:
      raise ValueError(
        f'Length of keys ({len(keys)}) does not match number of embeddings {embeddings.shape[0]}.'
      )

    current_size = self.size() if self._embeddings is not None else 0

    # Cast to float32 since dot product with float32 is 40-50x faster than float16 and 2.5x faster
    # than float64.
    if self._embeddings is None:
      self._embeddings = embeddings.astype(np.float32)
    else:
      self._embeddings = np.concatenate([self._embeddings, embeddings.astype(np.float32)], axis=0)

    row_indices = np.arange(current_size, current_size + len(keys), dtype=np.int32)
    new_key_to_label = pd.Series(row_indices, index=keys, dtype=np.int32)
    if self._key_to_index is not None:
      self._key_to_index = pd.concat([self._key_to_index, new_key_to_label])
    else:
      self._key_to_index = new_key_to_label

  @override
  def get(self, keys: Optional[Iterable[VectorKey]] = None) -> np.ndarray:
    assert (
      self._embeddings is not None and self._key_to_index is not None
    ), 'The vector store has no embeddings. Call load() or add() first.'
    if not keys:
      return self._embeddings
    locs = self._key_to_index.loc[cast(list[str], keys)]
    return self._embeddings.take(locs, axis=0)

  @override
  def topk(
    self, query: np.ndarray, k: int, keys: Optional[Iterable[VectorKey]] = None
  ) -> list[tuple[VectorKey, float]]:
    assert (
      self._embeddings is not None and self._key_to_index is not None
    ), 'The vector store has no embeddings. Call load() or add() first.'
    if keys is not None:
      row_indices = self._key_to_index.loc[cast(list[str], keys)]
      embeddings = self._embeddings.take(row_indices, axis=0)
      keys = list(keys)
    else:
      keys, embeddings = (
        cast(list[VectorKey], self._key_to_index.index.tolist()),
        self._embeddings,
      )

    query = query.astype(embeddings.dtype)
    similarities: np.ndarray = np.dot(embeddings, query).reshape(-1)
    k = min(k, len(similarities))

    # We do a partition + sort only top K to save time: O(n + klogk) instead of O(nlogn).
    indices = np.argpartition(similarities, -k)[-k:]
    # Indices sorted by value from largest to smallest.
    indices = indices[np.argsort(similarities[indices])][::-1]

    topk_similarities = similarities[indices]
    topk_keys = [keys[idx] for idx in indices]
    return list(zip(topk_keys, topk_similarities))
